import os
from tempfile import TemporaryDirectory

import pytest

from debian.substvars import Substvars, Substvar


class TestSubstvars:

    def test_substvars(self):
        # type: () -> None
        substvars = Substvars()

        assert substvars.substvars_path is None, None

        # add_dependency automatically creates variables
        assert 'misc:Recommends' not in substvars
        substvars.add_dependency('misc:Recommends', "foo (>= 1.0)")
        assert substvars['misc:Recommends'] == 'foo (>= 1.0)'
        # It can be appended to other variables
        substvars['foo'] = 'bar, golf'
        substvars.add_dependency('foo', 'dpkg (>= 1.20.0)')
        assert substvars['foo'] == 'bar, dpkg (>= 1.20.0), golf'
        # Exact duplicates are ignored
        substvars.add_dependency('foo', 'dpkg (>= 1.20.0)')
        assert substvars['foo'] == 'bar, dpkg (>= 1.20.0), golf'

        substvar = substvars.as_substvar['foo']
        assert substvar.assignment_operator == "="
        substvar.assignment_operator = "?="

        with pytest.raises(ValueError):
            # Only "=" and "?=" are allowed
            substvar.assignment_operator = 'golf'

        assert 'foo' in substvars
        del substvars['foo']
        assert not ('foo' in substvars)

    def test_save_raises(self):
        # type: () -> None
        s = Substvars()
        with pytest.raises(TypeError):
            # Should raise because it has no base file
            s.save()

    def test_save(self):
        # type: () -> None
        with TemporaryDirectory() as tmpdir:
            filename = os.path.join(tmpdir, "foo.substvars")
            # Obviously, this does not exist
            assert not os.path.exists(filename)
            with Substvars.load_from_path(filename, missing_ok=True) as svars:
                svars.add_dependency("misc:Depends", "bar (>= 1.0)")
                svars.as_substvar["foo"] = Substvar("anything goes", assignment_operator="?=")
                assert svars.substvars_path == filename
            assert os.path.exists(filename)

            with Substvars.load_from_path(filename) as svars:
                # Verify we can actually load the file we just wrote again
                assert svars['misc:Depends'] == "bar (>= 1.0)"
                assert svars.as_substvar["misc:Depends"].assignment_operator == "="
                assert svars['foo'] == "anything goes"
                assert svars.as_substvar["foo"].assignment_operator == "?="

    def test_equals(self):
        # type: () -> None
        foo_a = Substvar("foo", assignment_operator="=")
        foo_b = Substvar("foo", assignment_operator="=")
        foo_optional_a = Substvar("foo", assignment_operator="?=")
        foo_optional_b = Substvar("foo", assignment_operator="?=")
        assert foo_a == foo_b
        assert foo_optional_a == foo_optional_b

        assert foo_a != foo_optional_a
        assert foo_a != object()

        substvars_a = Substvars()
        substvars_b = Substvars()
        substvars_a["foo"] = "bar"
        substvars_b["foo"] = "bar"
        assert substvars_a == substvars_b
        assert substvars_a != object()




try:
    from typing import Dict, Any
except ImportError:
    pass

import pytest

from debian.tests.stubbed_arch_table import StubbedDpkgArchTable


@pytest.fixture(autouse=True)
def doctest_add_load_arch_table(doctest_namespace):
    # type: (Dict[str, Any]) -> None
    # Provide a custom namespace for doctests such that we can have them use
    # a custom environment. Use sparingly.
    # - For this to work, the doctests MUST NOT import the names listed here
    #   (as the import would overwrite the stub)
    doctest_namespace['DpkgArchTable'] = StubbedDpkgArchTable


# CRUFT: can be deleted once stretch/Python 3.5 support is dropped
#
# pytest caplog was introduced in pytest 3.3; being able to run the test suite
# on Debian stretch (Python 3.5, pytest 3.0.6) is desired. Doing
# an ad hoc version comparison for this is useful

from debian.debian_support import Version

installed_pytest_version = Version(pytest.__version__)
caplog_min_version = Version("3.3")

if installed_pytest_version < caplog_min_version:
    @pytest.fixture()
    def caplog(request):   # type: ignore
        pytest.skip("caplog capability not present")



#!/usr/bin/python
# vim: fileencoding=utf-8
#
# changelog.py -- Python module for Debian changelogs
# Copyright (C) 2006-7 James Westby <jw+debian@jameswestby.net>
# Copyright (C) 2008 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# The parsing code is based on that from dpkg which is:
# Copyright 1996 Ian Jackson
# Copyright 2005 Frank Lichtenheld <frank@lichtenheld.de>
# and licensed under the same license as above.

import logging
import os.path

import pytest

from debian import changelog
from debian import debian_support


try:
    # pylint: disable=unused-import
    from typing import (
        Any,
        IO,
        Optional,
        Text,
    )
except ImportError:
    # Missing types aren't important at runtime
    pass


def find_test_file(filename):
    # type: (str) -> str
    """ find a test file that is located within the test suite """
    return os.path.join(os.path.dirname(__file__), filename)


def open_utf8(filename, mode='r'):
    # type: (str, str) -> IO[Text]
    """Open a UTF-8 text file in text mode."""
    return open(filename, mode=mode, encoding='UTF-8')


class TestChangelog:

    def test_create_changelog(self):
        # type: () -> None
        with open(find_test_file('test_changelog')) as f:
            c = f.read()
        cl = changelog.Changelog(c)
        cs = str(cl)
        clines = c.split('\n')
        cslines = cs.split('\n')
        for i in range(len(clines)):
            assert clines[i] == cslines[i]
        assert len(clines) == len(cslines), "Different lengths"

    def test_create_changelog_single_block(self):
        # type: () -> None
        with open(find_test_file('test_changelog')) as f:
            c = f.read()
        cl = changelog.Changelog(c, max_blocks=1)
        cs = str(cl)
        assert cs == \
        """gnutls13 (1:1.4.1-1) unstable; urgency=HIGH

  [ James Westby ]
  * New upstream release. Closes: #123, #456,
    #789. LP: #1234, #2345,
    #3456
  * Remove the following patches as they are now included upstream:
    - 10_certtoolmanpage.diff
    - 15_fixcompilewarning.diff
    - 30_man_hyphen_*.patch
  * Link the API reference in /usr/share/gtk-doc/html as gnutls rather than
    gnutls-api so that devhelp can find it.

 -- Andreas Metzler <ametzler@debian.org>  Sat, 15 Jul 2006 11:11:08 +0200

"""

    def test_modify_changelog(self):
        # type: () -> None
        with open(find_test_file('test_modify_changelog1')) as f:
            c = f.read()
        cl = changelog.Changelog(c)
        cl.package = 'gnutls14'
        cl.version = '1:1.4.1-2'
        cl.distributions = 'experimental'
        cl.urgency = 'medium'
        cl.add_change('  * Add magic foo')
        cl.author = 'James Westby <jw+debian@jameswestby.net>'
        cl.date = 'Sat, 16 Jul 2008 11:11:08 -0200'
        with open(find_test_file('test_modify_changelog2')) as f:
            c = f.read()
        clines = c.split('\n')
        cslines = str(cl).split('\n')
        for i in range(len(clines)):
            assert clines[i] == cslines[i]
        assert len(clines) == len(cslines), "Different lengths"

    def test_preserve_initial_lines(self, caplog):
        # type: (pytest.LogCaptureFixture) -> None
        cl_text = b"""
THIS IS A LINE THAT SHOULD BE PRESERVED BUT IGNORED
haskell-src-exts (1.8.2-3) unstable; urgency=low

  * control: Use versioned Replaces: and Conflicts:

 -- Somebody <nobody@debian.org>  Wed, 05 May 2010 18:01:53 -0300
"""
        with caplog.at_level(logging.WARNING):
            cl = changelog.Changelog(cl_text)
        assert caplog.record_tuples == [(
            "debian.changelog",
            logging.WARNING,
            'Unexpected line while looking for first heading: '
            'THIS IS A LINE THAT SHOULD BE PRESERVED BUT IGNORED'
        )]

        assert cl_text == bytes(cl)

    def test_add_changelog_section(self):
        # type: () -> None
        with open(find_test_file('test_modify_changelog2')) as f:
            c = f.read()
        cl = changelog.Changelog(c)
        cl.new_block(package='gnutls14',
                version=debian_support.Version('1:1.4.1-3'),
                distributions='experimental',
                urgency='low',
                author='James Westby <jw+debian@jameswestby.net>')

        with pytest.raises(changelog.ChangelogCreateError):
            cl.__str__()

        cl.set_date('Sat, 16 Jul 2008 11:11:08 +0200')
        cl.add_change('')
        cl.add_change('  * Foo did not work, let us try bar')
        cl.add_change('')

        f = open(find_test_file('test_modify_changelog3'))
        c = f.read()
        f.close()
        clines = c.split('\n')
        cslines = str(cl).split('\n')
        for i in range(len(clines)):
            assert clines[i] == cslines[i]
        assert len(clines) == len(cslines), "Different lengths"

    def test_strange_changelogs(self):
        # type: () -> None
        """ Just opens and parses a strange changelog """
        with open(find_test_file('test_strange_changelog')) as f:
            c = f.read()
        cl = changelog.Changelog(c)

    def test_set_version_with_string(self):
        # type: () -> None
        with open(find_test_file('test_modify_changelog1')) as f:
            c1 = changelog.Changelog(f.read())
            f.seek(0)
            c2 = changelog.Changelog(f.read())
        c1.version = '1:2.3.5-2'
        c2.version = debian_support.Version('1:2.3.5-2')
        assert c1.version == c2.version
        assert c1.full_version == c2.full_version
        assert c1.epoch == c2.epoch
        assert c1.upstream_version == c2.upstream_version
        assert c1.debian_version == c2.debian_version

    def test_changelog_no_author(self):
        # type: () -> None
        cl_no_author = """gnutls13 (1:1.4.1-1) unstable; urgency=low

  * New upstream release.

 --
"""
        c1 = changelog.Changelog()
        c1.parse_changelog(cl_no_author, allow_empty_author=True)
        assert c1.author == None
        assert c1.date == None
        assert c1.package == "gnutls13"
        with pytest.raises(changelog.ChangelogCreateError):
            str(c1)
        assert c1._format(allow_missing_author=True) == cl_no_author
        c2 = changelog.Changelog()
        with pytest.raises(changelog.ChangelogParseError):
            c2.parse_changelog(cl_no_author)

    def test_magic_version_properties(self):
        # type: () -> None
        with open(find_test_file('test_changelog')) as f:
            c = changelog.Changelog(f)
        assert c.debian_version == '1'
        assert c.full_version == '1:1.4.1-1'
        assert c.upstream_version == '1.4.1'
        assert c.epoch == '1'
        assert str(c.version) == c.full_version

    def test_bugs_closed(self):
        # type: () -> None
        with open(find_test_file('test_changelog')) as f:
            c = iter(changelog.Changelog(f))
        # test bugs in a list
        block = next(c)
        assert block.bugs_closed == [123, 456, 789]
        assert block.lp_bugs_closed == [1234, 2345, 3456]
        # test bugs in parentheses
        block = next(c)
        assert block.bugs_closed == [375815]
        assert block.lp_bugs_closed == []

    def test_allow_full_stops_in_distribution(self):
        # type: () -> None
        with open(find_test_file('test_changelog_full_stops')) as f:
            c = changelog.Changelog(f)
        assert c.debian_version == None
        assert c.full_version == '1.2.3'
        assert str(c.version) == c.full_version

    def test_str_consistent(self):
        # type: () -> None
        # The parsing of the changelog (including the string representation)
        # should be consistent whether we give a single string, a list of
        # lines, or a file object to the Changelog initializer
        with open(find_test_file('test_changelog')) as f:
            cl_data = f.read()
            f.seek(0)
            c1 = changelog.Changelog(f)
        c2 = changelog.Changelog(cl_data)
        c3 = changelog.Changelog(cl_data.splitlines())
        for c in (c1, c2, c3):
            assert str(c) == cl_data

    def test_utf8_encoded_file_input(self):
        # type: () -> None
        f = open_utf8(find_test_file('test_changelog_unicode'))
        c = changelog.Changelog(f)
        f.close()
        u = str(c)
        expected_u = """haskell-src-exts (1.8.2-3) unstable; urgency=low

  * control: Use versioned Replaces: and Conflicts:

 -- Marco T\xfalio Gontijo e Silva <marcot@debian.org>  Wed, 05 May 2010 18:01:53 -0300

haskell-src-exts (1.8.2-2) unstable; urgency=low

  * debian/control: Rename -doc package.

 -- Marco T\xfalio Gontijo e Silva <marcot@debian.org>  Tue, 16 Mar 2010 10:59:48 -0300
"""
        assert u == expected_u
        assert bytes(c) == u.encode('utf-8')

    def test_unicode_object_input(self):
        # type: () -> None
        with open(find_test_file('test_changelog_unicode'), 'rb') as f:
            c_bytes = f.read()
        c_unicode = c_bytes.decode('utf-8')
        c = changelog.Changelog(c_unicode)
        assert str(c) == c_unicode
        assert bytes(c) == c_bytes

    def test_non_utf8_encoding(self):
        # type: () -> None
        with open(find_test_file('test_changelog_unicode'), 'rb') as f:
            c_bytes = f.read()
        c_unicode = c_bytes.decode('utf-8')
        c_latin1_str = c_unicode.encode('latin1')
        c = changelog.Changelog(c_latin1_str, encoding='latin1')
        assert str(c) == c_unicode
        assert bytes(c) == c_latin1_str
        for block in c:
            assert bytes(block) == str(block).encode('latin1')

    def test_malformed_date(self, caplog):
        # type: (pytest.LogCaptureFixture) -> None
        c_text = """package (1.0-1) codename; urgency=medium

  * minimal example reproducer of malformed date line

 -- John Smith <john.smith@example.com> Tue, 27 Sep 2016 14:08:04 -0600
 """
        # In strict mode, exceptions should be raised by the malformed entry
        with pytest.raises(changelog.ChangelogParseError):
            c = changelog.Changelog(c_text, strict=True)
        # In non-strict mode, warnings should be emitted by the malformed entry
        with caplog.at_level(logging.WARNING):
            c = changelog.Changelog(c_text, strict=False)
        assert len(c) == 1
        assert caplog.record_tuples == [(
            "debian.changelog",
            logging.WARNING,
            'Badly formatted trailer line:  '
            '-- John Smith <john.smith@example.com> '
            'Tue, 27 Sep 2016 14:08:04 -0600'
        )]

    def test_block_iterator(self):
        # type: () -> None
        with open(find_test_file('test_changelog')) as f:
            c = changelog.Changelog(f)
        assert [str(b) for b in c._blocks] == [str(b) for b in c]

    def test_block_access(self):
        # type: () -> None
        """ test random access to changelog entries """
        with open(find_test_file('test_changelog')) as f:
            c = changelog.Changelog(f)
        assert str(c[2].version) == '1.4.0-2', \
                         'access by sequence number'
        assert str(c['1.4.0-1'].version) == '1.4.0-1', \
                         'access by version string'
        assert str(c[debian_support.Version('1.3.5-1.1')].version) == \
                         '1.3.5-1.1', \
                         'access by Version object'

    def test_len(self):
        # type: () -> None
        with open(find_test_file('test_changelog')) as f:
            c = changelog.Changelog(f)
        assert len(c._blocks) == len(c)



#!/usr/bin/python3
# -*- coding: utf-8 -*- vim: fileencoding=utf-8 :

# Copyright (C) 2021 Niels Thykier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for format preserving deb822"""
import collections
import contextlib
import logging
import textwrap
from debian.deb822 import Deb822

import pytest

from debian._deb822_repro import (parse_deb822_file,
                                  AmbiguousDeb822FieldKeyError,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  LIST_COMMA_SEPARATED_INTERPRETATION,
                                  Interpretation,
                                  )
from debian._deb822_repro.parsing import Deb822KeyValuePairElement, Deb822ParsedTokenList, Deb822ParagraphElement, \
    Deb822FileElement, Deb822ParsedValueElement, LIST_UPLOADERS_INTERPRETATION
from debian._deb822_repro.tokens import Deb822Token, Deb822ErrorToken
from debian._deb822_repro._util import print_ast

try:
    from typing import Any, Iterator, Tuple
    from debian._deb822_repro.types import VE, ST
except ImportError:
    pass

RoundTripParseCase = collections.namedtuple('RoundTripParseCase',
                                            ['input',
                                             'is_valid_file',
                                             'error_element_count',
                                             'duplicate_fields',
                                             'paragraph_count',
                                             ])

# We use ¶ as "end of line" marker for two reasons in cases with optional whitespace:
# - to show that we have it when you debug the test case
# - to stop formatters from stripping it
#
# The marker is not required.  Consider to omit it if the test case does not
# involve trailing whitespace.
#
# NB: As a side-effect of the implementation, the tests strips '¶' unconditionally.
# Please another fancy glyph if you need to test non-standard characters.
ROUND_TRIP_CASES = [
    RoundTripParseCase(input='\n',
                       is_valid_file=False,
                       error_element_count=0,
                       duplicate_fields=False,
                       paragraph_count=0
                       ),
    RoundTripParseCase(input='A: b\n',
                       is_valid_file=True,
                       error_element_count=0,
                       duplicate_fields=False,
                       paragraph_count=1
                       ),
    RoundTripParseCase(input=textwrap.dedent('''\
                        Source: debhelper
                        # Trailing-whitespace
                        # Comment before a field
                        Build-Depends: po4a
                        # Ending with an empty field
                        Empty-Field:
                        '''),
                       is_valid_file=True,
                       error_element_count=0,
                       duplicate_fields=False,
                       paragraph_count=1
                       ),
    RoundTripParseCase(input=textwrap.dedent('''\
                        Source: debhelper
                        # Trailing-whitespace
                        # Comment before a field
                        Build-Depends: po4a

                        #  Comment about debhelper
                        Package: debhelper
                        Architecture: all
                        Depends: something
                        # We also depend on libdebhelper-perl
                                 libdebhelper-perl (= ${binary:Version})
                        Description: something
                         A long
                        \tand boring
                         description
                        # And a continuation line (also, inline comment)
                         .
                         Final remark
                        # Comment at the end of a paragraph plus multiple empty lines



                        # This paragraph contains a lot of trailing-whitespace cases, so we¶
                        # will be using the end of line marker through out this paragraph  ¶
                        Package: libdebhelper-perl¶
                        Priority:optional ¶
                        # Allowed for debian/control file
                        Empty-Field:¶
                        Section:   section   ¶
                        #   Field starting  with     a space + newline (special-case)¶
                        Depends:¶
                                 ${perl:Depends},¶
                        # Some people like the "leading comma" solution to dependencies¶
                        # so we should we have test case for that as well.¶
                        # (makes more sense when we start to parse the field as a dependency¶
                        # field)¶
                        Suggests: ¶
                                , something  ¶
                                , another¶
                        # Field that ends without a newline¶
                        Architecture: all¶
                        '''),
                       paragraph_count=3,
                       is_valid_file=True,
                       duplicate_fields=False,
                       error_element_count=0,
                       ),
    RoundTripParseCase(input=textwrap.dedent('''\
                        Source: debhelper
                        # Missing colon
                        Build-Depends po4a

                        #  Comment about debhelper
                        Package: debhelper
                        Depends: something
                        # We also depend on libdebhelper-perl
                                 libdebhelper-perl (= ${binary:Version})
                        Description: something
                         A long
                         and boring
                         description
                        # Missing the dot

                         Final remark


                        Package: libdebhelper-perl
                        '''),
                       paragraph_count=3,
                       is_valid_file=False,
                       duplicate_fields=False,
                       error_element_count=2,
                       ),
    RoundTripParseCase(input=textwrap.dedent('''\
                        A: b
                        B: c
                        # Duplicate field
                        A: b
                        '''),
                       is_valid_file=False,
                       error_element_count=0,
                       duplicate_fields=True,
                       paragraph_count=1
                       ),
    RoundTripParseCase(input=textwrap.dedent('''\
                    Is-Valid-Paragraph: yes

                    Is-Valid-Paragraph: Definitely not
                    Package: foo
                    Package: bar
                    Something-Else:
                    # Some comment
                     asd
                    Package: baz
                    Another-Field: foo
                    Package: again
                    I-Can-Haz-Package: ?
                    Package: yes
                    '''),
                       is_valid_file=False,
                       error_element_count=0,
                       duplicate_fields=True,
                       paragraph_count=2
                       ),
]



class TestFormatPreservingDeb822Parser:

    def test_round_trip_cases(self):
        # type: () -> None

        for i, parse_case in enumerate(ROUND_TRIP_CASES, start=1):
            c = str(i)
            case_input = parse_case.input.replace('¶', '')
            try:
                deb822_file = parse_deb822_file(case_input.splitlines(keepends=True),
                                                accept_files_with_duplicated_fields=True,
                                                accept_files_with_error_tokens=True,
                                                )
            except Exception:
                logging.info("Error while parsing case " + c)
                raise
            error_element_count = 0
            for token in deb822_file.iter_tokens():
                if isinstance(token, Deb822ErrorToken):
                    error_element_count += 1

            if parse_case.error_element_count > 0 or parse_case.duplicate_fields:
                with pytest.raises(ValueError):
                    # By default, we would reject this file.
                    parse_deb822_file(case_input.splitlines(keepends=True))
            else:
                # The field should be accepted without any errors by default
                parse_deb822_file(case_input.splitlines(keepends=True))

            paragraph_count = len(list(deb822_file))
            # Remember you can use _print_ast(deb822_file) if you need to debug the test cases.
            # A la
            #
            # if i in (3, 4):
            #   logging.info(f" ---  CASE {i} --- ")
            #   _print_ast(deb822_file)
            #   logging.info(f" ---  END CASE {i} --- ")
            assert parse_case.error_element_count == error_element_count, \
                "Correct number of error tokens for case " + c
            assert parse_case.paragraph_count == paragraph_count, \
                "Correct number of paragraphs parsed for case " + c
            assert parse_case.is_valid_file == deb822_file.is_valid_file, \
                "Verify deb822_file correctly determines whether the field is invalid" \
                " for case " + c
            assert case_input == deb822_file.convert_to_text(), \
                             "Input of case " + c + " is round trip safe"

            newline_normalized_by_omission = parse_deb822_file(
                case_input.splitlines(),
                accept_files_with_duplicated_fields=True,
                accept_files_with_error_tokens=True,
            )
            case_input_newline_normalized = case_input.replace("\r", "")
            if not case_input_newline_normalized.endswith("\n") and len(case_input_newline_normalized.splitlines()) > 1:
                case_input_newline_normalized += "\n"
            assert case_input_newline_normalized == \
                             newline_normalized_by_omission.convert_to_text(), \
                             "Input of case " + c + " is newline normalized round trip safe" \
                                                    " with newlines omitted"
            logging.info("Successfully passed case " + c)

    def test_deb822_emulation(self):
        # type: () -> None

        for i, parse_case in enumerate(ROUND_TRIP_CASES, start=1):
            if not parse_case.is_valid_file:
                continue
            c = str(i)
            case_input = parse_case.input.replace('¶', '')
            try:
                deb822_file = parse_deb822_file(case_input.splitlines(keepends=True))
            except Exception:
                logging.info("Error while parsing case " + c)
                raise
            deb822_paragraphs = list(Deb822.iter_paragraphs(case_input.splitlines()))

            for repro_paragraph, deb822_paragraph in zip(deb822_file, deb822_paragraphs):
                assert list(repro_paragraph) == list(deb822_paragraph), \
                    "Ensure keys are the same and in the correct order, case " + c
                # Use the key from Deb822 as it is compatible with the round safe version
                # (the reverse is not true typing wise)
                for k, ev in deb822_paragraph.items():
                    av = repro_paragraph[k]
                    assert av == ev, "Ensure value for " + k + " is the same, case " + c

    def test_regular_fields(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          # Comment for S-V
          Standards-Version: 1.2.3
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        source_paragraph = next(iter(deb822_file))
        assert "foo" == source_paragraph['Source']
        assert "1.2.3" == source_paragraph['Standards-Version']
        assert "no" == source_paragraph['Rules-Requires-Root']

        # Test setter and deletion while we are at it
        source_paragraph["Rules-Requires-Root"] = "binary-targets"
        source_paragraph["New-Field"] = "value"
        del source_paragraph["Standards-Version"]

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: binary-targets
          New-Field: value
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

        # As an alternative, we can also fix the problem if we discard comments
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))
        as_dict_discard_comments = source_paragraph.configured_view(
            preserve_field_comments_on_field_updates=False,
            auto_resolve_ambiguous_fields=False,
        )
        # Test setter and deletion while we are at it
        as_dict_discard_comments["Rules-Requires-Root"] = "binary-targets"
        as_dict_discard_comments["New-Field"] = "value"
        del as_dict_discard_comments["Standards-Version"]
        expected = textwrap.dedent('''\
          Source: foo
          Rules-Requires-Root: binary-targets
          New-Field: value
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while but discarded comments"

        source_paragraph['Multi-Line-Field-Space'] = textwrap.dedent('''\
        foo
         bar
        ''')
        source_paragraph['Multi-Line-Field-Tab'] = textwrap.dedent('''\
        foo
        \tbar
        ''')
        expected = textwrap.dedent('''\
          Source: foo
          Rules-Requires-Root: binary-targets
          New-Field: value
          Multi-Line-Field-Space: foo
           bar
          Multi-Line-Field-Tab: foo
          \tbar
          ''')
        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving space + tab"

    def test_empty_fields(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          Field: foo
          Empty-Field:''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        source_paragraph = next(iter(deb822_file))
        assert "" == source_paragraph['Empty-Field']
        source_paragraph['Another-Empty-Field'] = ""
        assert "" == source_paragraph['Another-Empty-Field']
        list_view = source_paragraph.as_interpreted_dict_view(LIST_SPACE_SEPARATED_INTERPRETATION)
        with list_view['Empty-Field'] as empty_field:
            assert not bool(empty_field)

        with list_view['Field'] as field:
            assert bool(field)
            field.clear()
            assert not bool(field)

        expected = textwrap.dedent('''\
          Source: foo
          Field:
          Empty-Field:
          Another-Empty-Field:
          ''')
        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked and generate a valid file"

    def test_empty_fields_reorder(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          Field: foo
          Empty-Field:''')
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))
        source_paragraph.order_last('Field')
        expected = textwrap.dedent('''\
          Source: foo
          Empty-Field:
          Field: foo
          ''')
        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked and generate a valid file"
        # Re-parse
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))
        source_paragraph.order_first('Empty-Field')
        expected = textwrap.dedent('''\
          Empty-Field:
          Source: foo
          Field: foo
          ''')
        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked and generate a valid file"

    def test_case_preservation(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          rules-requires-root: no
          # Comment for S-V
          Standards-Version: 1.2.3
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        source_paragraph = next(iter(deb822_file))
        assert "foo" == source_paragraph['Source']
        assert "1.2.3" == source_paragraph['Standards-Version']
        assert "no" == source_paragraph['Rules-Requires-Root']

        # Test setter and deletion while we are at it
        source_paragraph["Rules-Requires-Root"] = "binary-targets"
        source_paragraph["New-field"] = "value"
        del source_paragraph["Standards-Version"]

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          rules-requires-root: binary-targets
          New-field: value
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving case"

        # Repeat with duplicated fields
        original = textwrap.dedent('''\
          source: foo
          source: foo
          # Comment for RRR
          rules-requires-root: no
          rules-requires-root: no
          # Comment for S-V
          Standards-Version: 1.2.3
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True),
                                        accept_files_with_duplicated_fields=True,
                                        )

        source_paragraph = next(iter(deb822_file))
        assert "foo" == source_paragraph['Source']
        assert "1.2.3" == source_paragraph['Standards-Version']
        assert "no" == source_paragraph['Rules-Requires-Root']

        # Test setter and deletion while we are at it
        source_paragraph["Rules-Requires-Root"] = "binary-targets"
        source_paragraph["New-field"] = "value"
        del source_paragraph["Standards-Version"]

        expected = textwrap.dedent('''\
          source: foo
          source: foo
          # Comment for RRR
          rules-requires-root: binary-targets
          New-field: value
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving case"

    def test_preserve_field_order_on_mutation(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          Section: bar
          Priority: extra
          # Comment for RRR
          Rules-Requires-Root: no
          Build-Depends: debhelper-compat (= 10)
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        source_paragraph = next(iter(deb822_file))
        assert "foo" == source_paragraph['Source']

        source_paragraph["Rules-Requires-Root"] = "binary-targets"
        source_paragraph["section"] = "devel"
        source_paragraph["Priority"] = "optional"

        expected = textwrap.dedent('''\
          Source: foo
          Section: devel
          Priority: optional
          # Comment for RRR
          Rules-Requires-Root: binary-targets
          Build-Depends: debhelper-compat (= 10)
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving field order"

        # Again - this time with a paragraph containing duplicate fields
        original = textwrap.dedent('''\
          Source: foo
          Source: foo2
          Section: bar
          Section: baz
          Priority: extra
          # Comment for RRR
          Rules-Requires-Root: no
          Build-Depends: debhelper-compat (= 10)
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True),
                                        accept_files_with_duplicated_fields=True,
                                        )

        source_paragraph = next(iter(deb822_file))
        assert "foo" == source_paragraph['Source']

        source_paragraph["Rules-Requires-Root"] = "binary-targets"
        source_paragraph["section"] = "devel"
        source_paragraph["Priority"] = "optional"

        expected = textwrap.dedent('''\
          Source: foo
          Source: foo2
          Section: devel
          Priority: optional
          # Comment for RRR
          Rules-Requires-Root: binary-targets
          Build-Depends: debhelper-compat (= 10)
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving field order"

    def test_preserve_field_case_on_iter(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          secTion: bar
          PrIorIty: extra
          # Comment for RRR
          Rules-Requires-Root: no
          Build-Depends: debhelper-compat (= 10)
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        source_paragraph = next(iter(deb822_file))
        expected_keys = {
            'Source',
            'secTion',
            'PrIorIty',
            'Rules-Requires-Root',
            'Build-Depends'
        }
        actual_keys = set(source_paragraph.keys())

        assert expected_keys == actual_keys, \
            "Keys returned by iterations should have original case"

    def test_append_paragraph(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'bar'
        binary_paragraph['Description'] = 'Binary package bar'

        deb822_file.append(binary_paragraph)

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          Package: bar
          Description: Binary package bar
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_append_paragraph_existing_trailing_newline(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'bar'
        binary_paragraph['Description'] = 'Binary package bar'

        deb822_file.append(binary_paragraph)

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          Package: bar
          Description: Binary package bar
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_append_empty_paragraph(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()

        deb822_file.append(binary_paragraph)

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_append_tailing_comment(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          # Foo
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'bar'
        binary_paragraph['Description'] = 'Binary package bar'

        deb822_file.append(binary_paragraph)

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          # Foo

          Package: bar
          Description: Binary package bar
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_insert_paragraph(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'bar'
        binary_paragraph['Description'] = 'Binary package bar'

        deb822_file.insert(0, binary_paragraph)

        expected = textwrap.dedent('''\
          Package: bar
          Description: Binary package bar

          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

        # Insert after the existing paragraphs

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'blah'
        binary_paragraph['Description'] = 'Binary package blah'

        deb822_file.insert(5, binary_paragraph)

        expected = textwrap.dedent('''\
          Package: bar
          Description: Binary package bar

          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          Package: blah
          Description: Binary package blah
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_insert_paragraph_with_comments(self):
        # type: () -> None

        # Note that it is unspecified where the "Package: bar"-paragraph is
        # inserted relative to the "# Initial comment"-comment.  This test case
        # only asserts that it does not change unknowingly.

        original = textwrap.dedent('''\
          # Initial comment

          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          # Comment
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'bar'
        binary_paragraph['Description'] = 'Binary package bar'

        deb822_file.insert(0, binary_paragraph)

        expected = textwrap.dedent('''\
          Package: bar
          Description: Binary package bar

          # Initial comment

          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          # Comment
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

        # Insert after the existing paragraphs

        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'blah'
        binary_paragraph['Description'] = 'Binary package blah'

        deb822_file.insert(5, binary_paragraph)

        expected = textwrap.dedent('''\
          Package: bar
          Description: Binary package bar

          # Initial comment

          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          # Comment

          Package: blah
          Description: Binary package blah
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_insert_paragraph_in_empty_file(self):
        # type: () -> None

        deb822_file = Deb822FileElement.new_empty_file()
        binary_paragraph = Deb822ParagraphElement.new_empty_paragraph()
        binary_paragraph['Package'] = 'bar'
        binary_paragraph['Description'] = 'Binary package bar'
        # There is a special-case for idx == 0 and that should be well-behaved
        # for empty files too.
        deb822_file.insert(0, binary_paragraph)

        expected = textwrap.dedent('''\
          Package: bar
          Description: Binary package bar
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_remove_paragraph(self):
        # type: () -> None
        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          Package: bar
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = list(deb822_file)[1]
        assert 'bar' == binary_paragraph['Package']

        deb822_file.remove(binary_paragraph)

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

        # Verify that we can add another paragraph.
        deb822_file.append(Deb822ParagraphElement.from_dict({'Package': 'bloe'}))

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          Package: bloe
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Adding new paragraph should have worked"

        deb822_file.remove(list(deb822_file)[1])

        source_paragraph = list(deb822_file)[0]
        assert 'foo' == source_paragraph['Source']

        deb822_file.remove(source_paragraph)

        expected = textwrap.dedent('''\
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

        original = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          Package: bar

          # Comment

          Package: la
          ''')

        deb822_file = parse_deb822_file(original.splitlines(keepends=True))

        binary_paragraph = list(deb822_file)[1]
        assert 'bar' == binary_paragraph['Package']

        deb822_file.remove(binary_paragraph)

        expected = textwrap.dedent('''\
          Source: foo
          # Comment for RRR
          Rules-Requires-Root: no

          # Comment

          Package: la
          ''')

        assert expected == deb822_file.convert_to_text(), \
            "Mutation should have worked while preserving comments"

    def test_duplicate_fields(self):
        # type: () -> None

        original = textwrap.dedent('''\
        Source: foo
        # Comment for RRR
        Rules-Requires-Root: no
        # Comment for S-V
        Standards-Version: 1.2.3
        Rules-Requires-Root: binary-targets
        ''')
        # By default, the file is accepted
        deb822_file = parse_deb822_file(original.splitlines(keepends=True),
                                        accept_files_with_duplicated_fields=True,
                                        )

        with pytest.raises(ValueError):
            # But the parser should raise an error if explicitly requested
            parse_deb822_file(original.splitlines(keepends=True),
                              accept_files_with_error_tokens=True,
                              accept_files_with_duplicated_fields=False,
                              )

        source_paragraph = next(iter(deb822_file))
        as_dict = source_paragraph.configured_view(auto_resolve_ambiguous_fields=False)
        # Non-ambiguous fields are fine
        assert "foo" == as_dict['Source']
        assert "1.2.3" == as_dict['Standards-Version']
        # Contains doesn't raise a AmbiguousDeb822FieldKeyError
        assert 'Rules-Requires-Root' in as_dict
        with pytest.raises(AmbiguousDeb822FieldKeyError):
            v = as_dict['Rules-Requires-Root']
        as_dict_auto_resolve = source_paragraph.configured_view(auto_resolve_ambiguous_fields=True)
        assert "foo" == as_dict_auto_resolve['Source']
        assert "1.2.3" == as_dict_auto_resolve['Standards-Version']
        # Auto-resolution always takes the first field value
        assert "no" == as_dict_auto_resolve['Rules-Requires-Root']
        # It should be possible to "fix" the duplicate field by setting the field explicitly
        as_dict_auto_resolve['Rules-Requires-Root'] = as_dict_auto_resolve['Rules-Requires-Root']

        expected_fixed = original.replace('Rules-Requires-Root: binary-targets\n', '')
        assert expected_fixed == deb822_file.convert_to_text(), \
            "Fixed version should only have one Rules-Requires-Root field"

        # As an alternative, we can also fix the problem if we discard comments
        deb822_file = parse_deb822_file(original.splitlines(keepends=True),
                                        accept_files_with_duplicated_fields=True,
                                        )
        source_paragraph = next(iter(deb822_file))
        as_dict_discard_comments = source_paragraph.configured_view(
            preserve_field_comments_on_field_updates=False,
            auto_resolve_ambiguous_fields=False,
        )
        # First, ensure the reset succeeded
        with pytest.raises(AmbiguousDeb822FieldKeyError):
            v = as_dict_discard_comments['Rules-Requires-Root']
        as_dict_discard_comments["Rules-Requires-Root"] = "no"
        # Test setter and deletion while we are at it
        as_dict_discard_comments["New-Field"] = "value"
        del as_dict_discard_comments["Standards-Version"]
        as_dict_discard_comments['Source'] = 'bar'
        expected = textwrap.dedent('''\
        Source: bar
        Rules-Requires-Root: no
        New-Field: value
        ''')
        assert expected == deb822_file.convert_to_text(), \
            "Fixed version should only have one Rules-Requires-Root field"

    def test_sorting(self):
        # type: () -> None

        name_order = {
            f: i
            for i, f in enumerate([
                'source',
                'priority'
            ], start=0)
        }

        def key_func(field_name):
            # type: (str) -> Tuple[int, str]
            field_name_lower = field_name.lower()
            order = name_order.get(field_name_lower)
            if order is not None:
                return order, field_name_lower
            return len(name_order), field_name_lower

        # Note the lack of trailing newline is deliberate.
        # We want to ensure that sorting cannot trash the file even if the last
        # field does not end with a newline
        original_nodups = textwrap.dedent('''\
        Source: foo
        # Comment for RRR
        Rules-Requires-Root: no
        # Comment for S-V
        Standards-Version: 1.2.3
        Build-Depends: foo
        # With inline comment
                       bar
        Priority: optional''')

        sorted_nodups = textwrap.dedent('''\
        Source: foo
        Priority: optional
        Build-Depends: foo
        # With inline comment
                       bar
        # Comment for RRR
        Rules-Requires-Root: no
        # Comment for S-V
        Standards-Version: 1.2.3
        ''')

        original_with_dups = textwrap.dedent('''\
        Source: foo
        # Comment for RRR
        Rules-Requires-Root: no
        # Comment for S-V
        Standards-Version: 1.2.3
        Priority: optional
        # Comment for Second instance of RRR
        Rules-Requires-Root: binary-targets
        Build-Depends: foo
        # With inline comment
                       bar''')
        sorted_with_dups = textwrap.dedent('''\
        Source: foo
        Priority: optional
        Build-Depends: foo
        # With inline comment
                       bar
        # Comment for RRR
        Rules-Requires-Root: no
        # Comment for Second instance of RRR
        Rules-Requires-Root: binary-targets
        # Comment for S-V
        Standards-Version: 1.2.3
        ''')

        deb822_file_nodups = parse_deb822_file(original_nodups.splitlines(keepends=True))
        for paragraph in deb822_file_nodups:
            paragraph.sort_fields(key=key_func)

        assert sorted_nodups == deb822_file_nodups.convert_to_text(), \
            "Sorting without duplicated fields work"
        deb822_file_with_dups = parse_deb822_file(original_with_dups.splitlines(keepends=True),
                                                  accept_files_with_duplicated_fields=True,
                                                  )

        for paragraph in deb822_file_with_dups:
            paragraph.sort_fields(key=key_func)

        assert sorted_with_dups == deb822_file_with_dups.convert_to_text(), \
            "Sorting with duplicated fields work"

    def test_reorder_nodups(self):
        # type: () -> None
        content = textwrap.dedent("""
        Depends: bar
        Description: some-text
        Architecture: any
        Package: foo
        Recommends: baz
        """)
        deb822_file = parse_deb822_file(content.splitlines(keepends=True))
        paragraph = next(iter(deb822_file))

        # Verify the starting state
        assert list(paragraph.keys()) == \
                         ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']
        # no op
        paragraph.order_last('Recommends')
        assert list(paragraph.keys()) == \
                         ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']
        # no op
        paragraph.order_first('Depends')
        assert list(paragraph.keys()) == \
                         ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']

        paragraph.order_first('Package')
        assert list(paragraph.keys()) == \
                         ['Package', 'Depends', 'Description', 'Architecture', 'Recommends']

        paragraph.order_last('Description')
        assert list(paragraph.keys()) == \
                         ['Package', 'Depends', 'Architecture', 'Recommends', 'Description']

        paragraph.order_after('Recommends', 'Depends')
        assert list(paragraph.keys()) == \
                         ['Package', 'Depends', 'Recommends', 'Architecture', 'Description']

        paragraph.order_before('Architecture', 'Depends')
        assert list(paragraph.keys()) == \
                         ['Package', 'Architecture', 'Depends', 'Recommends', 'Description']

        with pytest.raises(ValueError):
            paragraph.order_after('Architecture', 'Architecture')
        with pytest.raises(ValueError):
            paragraph.order_before('Architecture', 'Architecture')
        with pytest.raises(KeyError):
            paragraph.order_before('Unknown-Field', 'Architecture')
        with pytest.raises(KeyError):
            paragraph.order_before('Architecture', 'Unknown-Field')

    def test_reorder_dups(self):
        # type: () -> None
        content = textwrap.dedent("""
        Depends: bar
        Description: some-text
        Description: some-more-text
        Architecture: any
        Package: foo
        Package: foo2
        Recommends: baz
        """)
        deb822_file = parse_deb822_file(content.splitlines(keepends=True),
                                        accept_files_with_duplicated_fields=True,
                                        )
        paragraph = next(iter(deb822_file))
        # Verify the starting state
        assert list(paragraph.keys()) == \
                         ['Depends', 'Description', 'Description', 'Architecture', 'Package',
                          'Package', 'Recommends']
        # no op
        paragraph.order_last('Recommends')
        assert list(paragraph.keys()) == \
                         ['Depends', 'Description', 'Description', 'Architecture', 'Package',
                          'Package', 'Recommends']
        # no op
        paragraph.order_first('Depends')
        assert list(paragraph.keys()) == \
                         ['Depends', 'Description', 'Description', 'Architecture', 'Package',
                          'Package', 'Recommends']

        paragraph.order_first('Package')
        assert list(paragraph.keys()) == \
                         ['Package', 'Package','Depends', 'Description', 'Description',
                          'Architecture', 'Recommends']

        # Relative order must be preserved in this case.
        assert paragraph["Package"] == "foo"
        assert paragraph[("Package", 0)] == "foo"
        assert paragraph[("Package", 1)] == "foo2"

        # Repeating order_first should be a noop
        paragraph.order_first('Package')
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Depends', 'Description', 'Description',
                          'Architecture', 'Recommends']

        # Relative order must be preserved in this case.
        assert paragraph["Package"] == "foo"
        assert paragraph[("Package", 0)] == "foo"
        assert paragraph[("Package", 1)] == "foo2"

        paragraph.order_last('Description')
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Depends', 'Architecture', 'Recommends',
                          'Description', 'Description']
        # Relative order must be preserved in this case.
        assert paragraph["Description"] == "some-text"
        assert paragraph[("Description", 0)] == "some-text"
        assert paragraph[("Description", 1)] == "some-more-text"

        # Repeating order_first should be a noop
        paragraph.order_last('Description')
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Depends', 'Architecture', 'Recommends',
                          'Description', 'Description']
        # Relative order must be preserved in this case.
        assert paragraph["Description"] == "some-text"
        assert paragraph[("Description", 0)] == "some-text"
        assert paragraph[("Description", 1)] == "some-more-text"

        paragraph.order_after('Recommends', 'Depends')
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Depends', 'Recommends', 'Architecture',
                          'Description', 'Description', ]

        paragraph.order_before('Architecture', 'Depends')
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]

        # And now, for some "fun stuff"

        # Lets move the last Description field in front of the first.
        paragraph.order_before(('Description', 1), ('Description', 0))
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]
        # Verify the relocation was successful
        assert paragraph["Description"] == "some-more-text"
        assert paragraph[("Description", 0)] == "some-more-text"
        assert paragraph[("Description", 1)] == "some-text"

        # And swap their relative positions again
        paragraph.order_after(('Description', 0), ('Description', 1))
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]
        # Verify the relocation was successful
        assert paragraph["Description"] == "some-text"
        assert paragraph[("Description", 0)] == "some-text"
        assert paragraph[("Description", 1)] == "some-more-text"

        # This should be a no-op
        paragraph.order_last(('Description', 1))
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]
        assert paragraph["Description"] == "some-text"
        assert paragraph[("Description", 0)] == "some-text"
        assert paragraph[("Description", 1)] == "some-more-text"

        # This should cause them to swap order
        paragraph.order_last(('Description', 0))
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]
        # Verify the relocation was successful
        assert paragraph["Description"] == "some-more-text"
        assert paragraph[("Description", 0)] == "some-more-text"
        assert paragraph[("Description", 1)] == "some-text"

        # This should be a no-op
        paragraph.order_first(('Package', 0))
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]

        # Relative order must be preserved in this case.
        assert paragraph["Package"] == "foo"
        assert paragraph[("Package", 0)] == "foo"
        assert paragraph[("Package", 1)] == "foo2"

        # This should cause them to swap order
        paragraph.order_first(('Package', 1))
        assert list(paragraph.keys()) == \
                         ['Package', 'Package', 'Architecture', 'Depends', 'Recommends',
                          'Description', 'Description', ]

        # Verify the relocation was successful
        assert paragraph["Package"] == "foo2"
        assert paragraph[("Package", 0)] == "foo2"
        assert paragraph[("Package", 1)] == "foo"

        with pytest.raises(ValueError):
            paragraph.order_after('Architecture', 'Architecture')
        with pytest.raises(ValueError):
            paragraph.order_before('Architecture', 'Architecture')
        with pytest.raises(KeyError):
            paragraph.order_before('Unknown-Field', 'Architecture')
        with pytest.raises(KeyError):
            paragraph.order_before('Architecture', 'Unknown-Field')

    def test_interpretation(self):
        # type: () -> None

        original = textwrap.dedent('''\
        Package: foo
        Architecture: amd64  i386
        # Also on kfreebsd
          kfreebsd-amd64  kfreebsd-i386
        # With leading comma :)
        Some-Comma-List: , a,  b , c d, e
        Multiline-Comma-List: some
        # Invisible drive by comment
          , fun ,with
          multi-line
        # With a comment inside it for added fun
          values
          ,
        # Also an invisible comment here
          separated by
           ,commas
        # Comments in final value
             >:)
        Uploaders: Someone <nobody@example.org>,
        # Comment that should not be visible in the list
          Margrete, I, Ruler <1@margrete.dk>
        # Some remark that should not be visible in the list even though
        # the comma is on the other side
          ,
          Margrete, II, Queen
        # We could list additional names here
          <2@margrete.dk>
        ''')
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))

        @contextlib.contextmanager
        def _field_mutation_test(
                kvpair,           # type: Deb822KeyValuePairElement
                interpretation,   # type: Interpretation[Deb822ParsedTokenList[VE, ST]]
                expected_output,  # type: str
                ):
            # type: (...) -> Iterator[Deb822ParsedTokenList[VE, ST]]
            original_value_element = kvpair.value_element
            with kvpair.interpret_as(interpretation) as value_list:
                yield value_list

            # We always match without the field comment to keep things simple.
            actual = kvpair.field_name + ":" + kvpair.value_element.convert_to_text()
            try:
                assert expected_output == actual
            except AssertionError:
                logging.info(" -- Debugging aid - START of AST for generated value --")
                print_ast(kvpair)
                logging.info(" -- Debugging aid - END of AST for generated value --")
                raise
            # Reset of value
            kvpair.value_element = original_value_element
            assert original == deb822_file.convert_to_text()

        arch_kvpair = source_paragraph.get_kvpair_element('Architecture')
        comma_list_kvpair = source_paragraph.get_kvpair_element('Some-Comma-List')
        multiline_comma_list_kvpair = source_paragraph.get_kvpair_element('Multiline-Comma-List')
        uploaders_kvpair = source_paragraph.get_kvpair_element('Uploaders')
        assert arch_kvpair is not None and comma_list_kvpair is not None \
            and multiline_comma_list_kvpair is not None and uploaders_kvpair is not None
        archs = arch_kvpair.interpret_as(LIST_SPACE_SEPARATED_INTERPRETATION)
        comma_list_misread = comma_list_kvpair.interpret_as(
            LIST_SPACE_SEPARATED_INTERPRETATION
        )
        assert ['amd64', 'i386', 'kfreebsd-amd64', 'kfreebsd-i386'] == \
                         list(archs)
        assert [',', 'a,', 'b', ',', 'c', 'd,', 'e'] == \
                         list(comma_list_misread)

        comma_list_correctly_read = comma_list_kvpair.interpret_as(
            LIST_COMMA_SEPARATED_INTERPRETATION
        )
        ml_comma_list = multiline_comma_list_kvpair.interpret_as(
            LIST_COMMA_SEPARATED_INTERPRETATION
        )
        ml_comma_list_w_comments = multiline_comma_list_kvpair.interpret_as(
            LIST_COMMA_SEPARATED_INTERPRETATION,
            discard_comments_on_read=False
        )
        uploaders_list = uploaders_kvpair.interpret_as(
            LIST_UPLOADERS_INTERPRETATION,
        )
        uploaders_list_with_comments = uploaders_kvpair.interpret_as(
            LIST_UPLOADERS_INTERPRETATION,
            discard_comments_on_read=False
        )

        assert list(comma_list_correctly_read) == ['a', 'b', 'c d', 'e']

        assert list(ml_comma_list) == [
            "some",
            "fun",
            "with\n  multi-line\n  values",
            "separated by",
            "commas\n     >:)"
        ]

        assert list(ml_comma_list_w_comments) == [
            "some",
            "fun",
            "with\n  multi-line\n# With a comment inside it for added fun\n  values",
            "separated by",
            "commas\n# Comments in final value\n     >:)"
        ]

        assert list(uploaders_list) == [
            "Someone <nobody@example.org>",
            "Margrete, I, Ruler <1@margrete.dk>",
            "Margrete, II, Queen\n  <2@margrete.dk>",
        ]

        assert list(uploaders_list_with_comments) == [
            "Someone <nobody@example.org>",
            "Margrete, I, Ruler <1@margrete.dk>",
            "Margrete, II, Queen\n# We could list additional names here\n  <2@margrete.dk>",
        ]


        # Interpretation must not change the content
        assert original == deb822_file.convert_to_text()

        # But we can choose to modify the content
        expected_result = 'Some-Comma-List: , a,  b , c d, e, f,g,\n'
        with _field_mutation_test(comma_list_kvpair,
                                  LIST_COMMA_SEPARATED_INTERPRETATION,
                                  expected_result) as comma_list:
            comma_list.no_reformatting_when_finished()
            comma_list.append('f')
            # We can also omit the space after a separator
            comma_list.append_separator(space_after_separator=False)
            comma_list.append('g')
            comma_list.append_separator(space_after_separator=False)

        # ... and this time we reformat to make it look nicer
        expected_result = textwrap.dedent('''\
            Some-Comma-List: a,
                             c d,
                             e,
            # Something important about "f"
            #
            # ... that spans multiple lines    ¶
                             f,
        ''').replace('¶', '')
        with _field_mutation_test(comma_list_kvpair,
                                  LIST_COMMA_SEPARATED_INTERPRETATION,
                                  expected_result) as comma_list:
            comma_list.reformat_when_finished()
            comma_list.append_comment('Something important about "f"')
            comma_list.append_comment('')
            # We can control spacing by explicitly using "#" and "\n"
            comma_list.append_comment('# ... that spans multiple lines    \n')
            comma_list.append('f')
            comma_list.remove('b')

        # If we choose the wrong type of interpretation, the result should still be a valid Deb822 file
        # (even if the contents gets a bit wrong).
        expected_result = textwrap.dedent('''\
             Some-Comma-List: ,
                              a,
                              b
                              ,
                              c
                              d,
                              e
                              f
             ''')
        with _field_mutation_test(comma_list_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as comma_list_misread:
            comma_list_misread.reformat_when_finished()
            comma_list_misread.append('f')

        # This method also preserves existing comments
        expected_result = textwrap.dedent('''\
             Architecture: amd64  i386
             # Also on kfreebsd
               kfreebsd-amd64  kfreebsd-i386
             # And now on hurd
              hurd-amd64
              hurd-i386
             ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            arch_list.append_comment("And now on hurd")
            arch_list.append('hurd-amd64')
            arch_list.append_newline()
            arch_list.append('hurd-i386')

        # ... removals and comments
        expected_result = textwrap.dedent('''\
             Architecture: amd64  linux-x32
             # And now on hurd
              hurd-amd64
                 ''')

        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            arch_list.append_comment("And now on hurd")
            arch_list.append('hurd-amd64')
            arch_list.remove('kfreebsd-amd64')
            arch_list.remove('kfreebsd-i386')
            arch_list.replace('i386', 'linux-x32')

        # Reformatting will also preserve comments
        expected_result = textwrap.dedent('''\
             Architecture: amd64
                           i386
             # Also on kfreebsd
                           kfreebsd-amd64
                           kfreebsd-i386
             # And now on hurd
                           hurd-amd64
                           hurd-i386
             ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.reformat_when_finished()
            arch_list.append_newline()
            arch_list.append_comment("And now on hurd")
            arch_list.append('hurd-amd64')
            arch_list.append('hurd-i386')

        # Test removals of first and last value
        expected_result = textwrap.dedent('''\
            Architecture: i386
            # Also on kfreebsd
              kfreebsd-amd64¶
                 ''').replace('¶', '')

        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            arch_list.remove('amd64')
            arch_list.remove('kfreebsd-i386')

        # Same, just using value references
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            for value_ref in arch_list.iter_value_references():
                if value_ref.value in ('amd64', 'kfreebsd-i386'):
                    value_ref.remove()
                    # Ensure that a second call fails
                    with pytest.raises(RuntimeError):
                        value_ref.remove()

                    # As does attempting to fetch the value
                    with pytest.raises(RuntimeError):
                        _ = value_ref.value

                    # As does attempting to mutate the value
                    with pytest.raises(RuntimeError):
                        value_ref.value = "foo"

        expected_result = textwrap.dedent('''\
            Architecture: linux-amd64  linux-i386
            # Also on kfreebsd
              kfreebsd-amd64  kfreebsd-i386
        ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            for value_ref in arch_list.iter_value_references():
                value = value_ref.value
                if '-' not in value:
                    value_ref.value = 'linux-' + value

        # Test removal of first line without comment will hoist up the next line
        # - note eventually we might support keeping the comment by doing a
        #   "\n# ...\n value".
        expected_result = textwrap.dedent('''\
            Architecture: kfreebsd-amd64  kfreebsd-i386
                 ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            arch_list.remove('amd64')
            arch_list.remove('i386')

        # Test removal of first line without comment will hoist up the next line
        # This is only similar to the previous test case because we have not
        # made the previous case preserve comments
        expected_result = textwrap.dedent('''\
            Architecture: hurd-amd64
                 ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            # Delete kfreebsd first (which will remove the comment)
            arch_list.remove('kfreebsd-i386')
            arch_list.remove('kfreebsd-amd64')
            arch_list.append_newline()
            arch_list.append('hurd-amd64')
            arch_list.remove('amd64')
            arch_list.remove('i386')

        # Test deletion of the last value, which will clear the field
        expected_result = textwrap.dedent('''\
            Architecture: hurd-amd64
                 ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.no_reformatting_when_finished()
            arch_list.append_comment("This will not appear in the output")
            assert arch_list
            arch_list.remove('kfreebsd-i386')
            arch_list.remove('kfreebsd-amd64')
            arch_list.remove('amd64')
            arch_list.remove('i386')
            # Field should be cleared now.
            assert not arch_list
            # Add a value (as leaving the field empty would raise an error
            # on leaving the with-statement)
            arch_list.append('hurd-amd64')

        # Test sorting of the field
        expected_result = textwrap.dedent('''\
                Architecture: amd64
                              hurd-amd64
                              hurd-i386
                              i386
                # Also on kfreebsd
                              kfreebsd-amd64
                              kfreebsd-i386
                              ppc64el
                 ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            # Sort does not promise a "nice" output, hench reformatting
            arch_list.reformat_when_finished()
            # Add a few extra as the field is "almost" sorted already.
            arch_list.append('ppc64el')
            arch_list.append('hurd-i386')
            arch_list.append('hurd-amd64')
            arch_list.sort()

        # Test sorting of the field with key-func
        expected_result = textwrap.dedent('''\
                Architecture: amd64
                              i386
                              ppc64el
                # Also on kfreebsd
                              kfreebsd-amd64
                              kfreebsd-i386
                # Also on hurd
                              hurd-amd64
                              hurd-i386
                 ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            # Sort does not promise a "nice" output, hench reformatting
            arch_list.reformat_when_finished()
            # Add a few extra as the field is "almost" sorted already.
            arch_list.append('ppc64el')
            arch_list.append('hurd-i386')
            arch_list.append_comment('Also on hurd')
            arch_list.append('hurd-amd64')
            order = {
                'linux': 0,
                'kfreebsd': 1,
                'hurd': 2,
            }

            def _key_func(v):
                # type: (str) -> Any
                if '-' in v:
                    ov = order.get(v.split('-')[0])
                    if ov is None:
                        ov = 0
                else:
                    ov = 0
                return ov, v

            arch_list.sort(key=_key_func)

    def test_interpretation_tab_preservation(self):
        # type: () -> None

        original = textwrap.dedent('''\
        Package: foo
        Architecture: amd64  i386
          kfreebsd-amd64  kfreebsd-i386
        Build-Depends: debhelper-compat (= 12)
        \t , foo
        ''')
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))

        arch_kvpair = source_paragraph.get_kvpair_element('Architecture')
        bd_kvpair = source_paragraph.get_kvpair_element('Build-Depends')
        assert arch_kvpair is not None and bd_kvpair is not None

        @contextlib.contextmanager
        def _field_mutation_test(
                kvpair,           # type: Deb822KeyValuePairElement
                interpretation,   # type: Interpretation[Deb822ParsedTokenList[VE, ST]]
                expected_output,  # type: str
                ):
            # type: (...) -> Iterator[Deb822ParsedTokenList[VE, ST]]
            original_value_element = kvpair.value_element
            with kvpair.interpret_as(interpretation) as value_list:
                yield value_list

            # We always match without the field comment to keep things simple.
            actual = kvpair.field_name + ":" + kvpair.value_element.convert_to_text()
            try:
                assert expected_output == actual
            except AssertionError:
                logging.info(" -- Debugging aid - START of AST for generated value --")
                print_ast(kvpair)
                logging.info(" -- Debugging aid - END of AST for generated value --")
                raise
            # Reset of value
            kvpair.value_element = original_value_element
            assert original == deb822_file.convert_to_text()

        # With reformatting - should use space
        expected_result = textwrap.dedent('''\
                  Architecture: amd64
                                i386
                                kfreebsd-amd64
                                kfreebsd-i386
                                hurd-amd64
                                hurd-i386
                   ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.reformat_when_finished()
            arch_list.append('hurd-amd64')
            arch_list.append('hurd-i386')

        # Without reformatting - should use space
        expected_result = textwrap.dedent('''\
                  Architecture: amd64  i386
                    kfreebsd-amd64  kfreebsd-i386
                   hurd-amd64 hurd-i386
                   ''')
        with _field_mutation_test(arch_kvpair,
                                  LIST_SPACE_SEPARATED_INTERPRETATION,
                                  expected_result) as arch_list:
            arch_list.append_newline()
            arch_list.append('hurd-amd64')
            arch_list.append('hurd-i386')

        # With reformatting - the default formatter uses space
        expected_result = textwrap.dedent('''\
                          Build-Depends: debhelper-compat (= 12),
                                         foo,
                                         bar (>= 1.0~),
                           ''')
        with _field_mutation_test(bd_kvpair,
                                  LIST_COMMA_SEPARATED_INTERPRETATION,
                                  expected_result) as bd_list:
            bd_list.reformat_when_finished()
            bd_list.append('bar (>= 1.0~)')

        # Without reformatting - should use tab
        expected_result = textwrap.dedent('''\
                  Build-Depends: debhelper-compat (= 12)
                  \t , foo
                  \t, bar (>= 1.0~)
                   ''')
        with _field_mutation_test(bd_kvpair,
                                  LIST_COMMA_SEPARATED_INTERPRETATION,
                                  expected_result) as bd_list:
            bd_list.append_newline()
            bd_list.append('bar (>= 1.0~)')

    def test_mutate_field_preserves_whitespace(self):
        # type: () -> None

        original = textwrap.dedent('''\
        Package: foo
        Build-Depends:
         debhelper-compat (= 11),
         uuid-dev
        ''')
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))
        source_paragraph['Build-Depends'] = source_paragraph['Build-Depends']
        assert original == deb822_file.convert_to_text()

        original = textwrap.dedent('''\
        Package: foo
        Build-Depends: 
         debhelper-compat (= 11),
         uuid-dev
        ''')
        deb822_file = parse_deb822_file(original.splitlines(keepends=True))
        source_paragraph = next(iter(deb822_file))
        source_paragraph['Build-Depends'] = ' \n debhelper-compat (= 11),\n uuid-dev'
        assert original == deb822_file.convert_to_text()



#! /usr/bin/python
## vim: fileencoding=utf-8

# Copyright (C) 2006 Adeodato Simó <dato@net.com.org.es>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from collections import namedtuple
import email.utils
import io
import logging
import os
import os.path
import pickle
import re
import subprocess
import sys
import tempfile
import textwrap
import warnings

import pytest

try:
    # skip tests that require apt_pkg if it is not available
    import apt_pkg
    _have_apt_pkg = True
except (ImportError, AttributeError):
    _have_apt_pkg = False

from debian import deb822
from debian.debian_support import Version


try:
    # pylint: disable=unused-import
    from typing import (
        Any,
        Callable,
        Dict,
        Generator,
        IO,
        List,
        Optional,
        Union,
        Text,
        Tuple,
        Type,
        TypeVar,
    )
except ImportError:
    # Missing types aren't important at runtime
    TYPE_CHECKING = False

    # Fake some definitions
    if not TYPE_CHECKING:
        TypeVar = lambda t: None


# Only run tests that rely on the gpgv signature validation executable if
# it is installed
#
# TODO: For portability, should we use shutil.which()? and set
# deb822.GPGV_EXECUTABLE to match that?
_have_gpgv = os.path.exists('/usr/bin/gpgv')


# Deterministic tests are good; automatically skipping tests because optional
# dependencies are not available is a way of accidentally missing problems.
# Here, we control whether missing dependencies result in skipping tests
# or if instead, missing dependencies cause test failures.
#
# FORBID_MISSING_APT_PKG:
#   any non-empty value for the environment variable FORBID_MISSING_APT_PKG
#   will mean that tests fail if apt_pkg (from python-apt) can't be found
#
# FORBID_MISSING_GPGV:
#   any non-empty value for the environment variable FORBID_MISSING_GPGV
#   will mean that tests fail if the gpgv program can't be found
FORBID_MISSING_APT_PKG = os.environ.get("FORBID_MISSING_APT_PKG", None)
FORBID_MISSING_GPGV = os.environ.get("FORBID_MISSING_GPGV", None)


UNPARSED_PACKAGE = '''\
Package: mutt
Priority: standard
Section: mail
Installed-Size: 4471
Maintainer: Adeodato Simó <dato@net.com.org.es>
Architecture: i386
Version: 1.5.12-1
Replaces: mutt-utf8
Provides: mail-reader, imap-client
Depends: libc6 (>= 2.3.6-6), libdb4.4, libgnutls13 (>= 1.4.0-0), libidn11 (>= 0.5.18), libncursesw5 (>= 5.4-5), libsasl2 (>= 2.1.19.dfsg1), exim4 | mail-transport-agent
Recommends: locales, mime-support
Suggests: urlview, aspell | ispell, gnupg, mixmaster, openssl, ca-certificates
Conflicts: mutt-utf8
Filename: pool/main/m/mutt/mutt_1.5.12-1_i386.deb
Size: 1799444
MD5sum: d4a9e124beea99d8124c8e8543d22e9a
SHA1: 5e3c295a921c287cf7cb3944f3efdcf18dd6701a
SHA256: 02853602efe21d77cd88056a4e2a4350f298bcab3d895f5f9ae02aacad81442b
Description: text-based mailreader supporting MIME, GPG, PGP and threading
 Mutt is a sophisticated text-based Mail User Agent. Some highlights:
 .
  * MIME support (including RFC1522 encoding/decoding of 8-bit message
    headers and UTF-8 support).
  * PGP/MIME support (RFC 2015).
  * Advanced IMAP client supporting SSL encryption and SASL authentication.
  * POP3 support.
  * Mailbox threading (both strict and non-strict).
  * Default keybindings are much like ELM.
  * Keybindings are configurable; Mush and PINE-like ones are provided as
    examples.
  * Handles MMDF, MH and Maildir in addition to regular mbox format.
  * Messages may be (indefinitely) postponed.
  * Colour support.
  * Highly configurable through easy but powerful rc file.
Tag: interface::text-mode, made-of::lang:c, mail::imap, mail::pop, mail::user-agent, protocol::imap, protocol::ipv6, protocol::pop, protocol::ssl, role::sw:client, uitoolkit::ncurses, use::editing, works-with::mail
Task: mail-server
'''
        

PARSED_PACKAGE = deb822.Deb822Dict([
    ('Package', 'mutt'),
    ('Priority', 'standard'),
    ('Section', 'mail'),
    ('Installed-Size', '4471'),
    ('Maintainer', 'Adeodato Simó <dato@net.com.org.es>'),
    ('Architecture', 'i386'),
    ('Version', '1.5.12-1'),
    ('Replaces', 'mutt-utf8'),
    ('Provides', 'mail-reader, imap-client'),
    ('Depends', 'libc6 (>= 2.3.6-6), libdb4.4, libgnutls13 (>= 1.4.0-0), libidn11 (>= 0.5.18), libncursesw5 (>= 5.4-5), libsasl2 (>= 2.1.19.dfsg1), exim4 | mail-transport-agent'),
    ('Recommends', 'locales, mime-support'),
    ('Suggests', 'urlview, aspell | ispell, gnupg, mixmaster, openssl, ca-certificates'),
    ('Conflicts', 'mutt-utf8'),
    ('Filename', 'pool/main/m/mutt/mutt_1.5.12-1_i386.deb'),
    ('Size', '1799444'),
    ('MD5sum', 'd4a9e124beea99d8124c8e8543d22e9a'),
    ('SHA1', '5e3c295a921c287cf7cb3944f3efdcf18dd6701a'),
    ('SHA256', '02853602efe21d77cd88056a4e2a4350f298bcab3d895f5f9ae02aacad81442b'),
    ('Description', '''text-based mailreader supporting MIME, GPG, PGP and threading
 Mutt is a sophisticated text-based Mail User Agent. Some highlights:
 .
  * MIME support (including RFC1522 encoding/decoding of 8-bit message
    headers and UTF-8 support).
  * PGP/MIME support (RFC 2015).
  * Advanced IMAP client supporting SSL encryption and SASL authentication.
  * POP3 support.
  * Mailbox threading (both strict and non-strict).
  * Default keybindings are much like ELM.
  * Keybindings are configurable; Mush and PINE-like ones are provided as
    examples.
  * Handles MMDF, MH and Maildir in addition to regular mbox format.
  * Messages may be (indefinitely) postponed.
  * Colour support.
  * Highly configurable through easy but powerful rc file.'''),
    ('Tag', 'interface::text-mode, made-of::lang:c, mail::imap, mail::pop, mail::user-agent, protocol::imap, protocol::ipv6, protocol::pop, protocol::ssl, role::sw:client, uitoolkit::ncurses, use::editing, works-with::mail'),
    ('Task', 'mail-server'), ])


GPG_SIGNED = [ '''\
-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA1

%s

-----BEGIN PGP SIGNATURE-----
Version: GnuPG v1.4.3 (GNU/Linux)
Comment: Signed by Adeodato Simó <dato@net.com.org.es>

iEYEARECAAYFAkRqYxkACgkQgyNlRdHEGIKccQCgnnUgfwYjQ7xd3zGGS2y5cXKt
CcYAoOLYDF5G1h3oR1iDNyeCI6hRW03S
=Um8T
-----END PGP SIGNATURE-----
''', '''\
-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA1

%s

-----BEGIN PGP SIGNATURE-----

iEYEARECAAYFAkRqYxkACgkQgyNlRdHEGIKccQCgnnUgfwYjQ7xd3zGGS2y5cXKt
CcYAoOLYDF5G1h3oR1iDNyeCI6hRW03S
=Um8T
-----END PGP SIGNATURE-----
''',
    ]


CHANGES_FILE = '''\
Format: 1.7
Date: Fri, 28 Dec 2007 17:08:48 +0100
Source: bzr-gtk
Binary: bzr-gtk
Architecture: source all
Version: 0.93.0-2
Distribution: unstable
Urgency: low
Maintainer: Debian Bazaar Maintainers <pkg-bazaar-maint@lists.alioth.debian.org>
Changed-By: Chris Lamb <chris@chris-lamb.co.uk>
Description:
 bzr-gtk    - provides graphical interfaces to Bazaar (bzr) version control
Closes: 440354 456438
Changes:
 bzr-gtk (0.93.0-2) unstable; urgency=low
 .
   [ Chris Lamb ]
   * Add patch for unclosed progress window. (Closes: #440354)
     Patch by Jean-François Fortin Tam <jeff@ecchi.ca>
   * Fix broken icons in .desktop files (Closes: #456438).
Files:
 0fd797f4138a9d4fdeb8c30597d46bc9 1003 python optional bzr-gtk_0.93.0-2.dsc
 d9523676ae75c4ced299689456f252f4 3860 python optional bzr-gtk_0.93.0-2.diff.gz
 8960459940314b21019dedd5519b47a5 168544 python optional bzr-gtk_0.93.0-2_all.deb
'''

CHECKSUM_CHANGES_FILE = '''\
Format: 1.8
Date: Wed, 30 Apr 2008 23:58:24 -0600
Source: python-debian
Binary: python-debian
Architecture: source all
Version: 0.1.10
Distribution: unstable
Urgency: low
Maintainer: Debian python-debian Maintainers <pkg-python-debian-maint@lists.alioth.debian.org>
Changed-By: John Wright <jsw@debian.org>
Description:
 python-debian - Python modules to work with Debian-related data formats
Closes: 473254 473259
Changes:
 python-debian (0.1.10) unstable; urgency=low
 .
   * debian_bundle/deb822.py, tests/test_deb822.py:
     - Do not cache _CaseInsensitiveString objects, since it causes case
       preservation issues under certain circumstances (Closes: #473254)
     - Add a test case
   * debian_bundle/deb822.py:
     - Add support for fixed-length subfields in multivalued fields.  I updated
       the Release and PdiffIndex classes to use this.  The default behavior for
       Release is that of apt-ftparchive, just because it's simpler.  Changing
       the behavior to resemble dak requires simply setting the
       size_field_behavior attribute to 'dak'.  (Ideally, deb822 would detect
       which behavior to use if given an actual Release file as input, but this
       is not implemented yet.)  (Closes: #473259)
     - Add support for Checksums-{Sha1,Sha256} multivalued fields in Dsc and
       Changes classes
   * debian/control:
     - "python" --> "Python" in the Description field
     - Change the section to "python"
Checksums-Sha1:
 d12d7c95563397ec37c0d877486367b409d849f5 1117 python-debian_0.1.10.dsc
 19efe23f688fb7f2b20f33d563146330064ab1fa 109573 python-debian_0.1.10.tar.gz
 22ff71048921a788ad9d90f9579c6667e6b3de3a 44260 python-debian_0.1.10_all.deb
Checksums-Sha256:
 aae63dfb18190558af8e71118813dd6a11157c6fd92fdc9b5c3ac370daefe5e1 1117 python-debian_0.1.10.dsc
 d297c07395ffa0c4a35932b58e9c2be541e8a91a83ce762d82a8474c4fc96139 109573 python-debian_0.1.10.tar.gz
 4c73727b6438d9ba60aeb5e314e2d8523f021da508405dc54317ad2b392834ee 44260 python-debian_0.1.10_all.deb
Files:
 469202dfd24d55a932af717c6377ee59 1117 python optional python-debian_0.1.10.dsc
 4857552b0156fdd4fa99d21ec131d3d2 109573 python optional python-debian_0.1.10.tar.gz
 81864d535c326c082de3763969c18be6 44260 python optional python-debian_0.1.10_all.deb
'''

SIGNED_CHECKSUM_CHANGES_FILE = '''\
-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA1

%s
-----BEGIN PGP SIGNATURE-----
Version: GnuPG v1.4.6 (GNU/Linux)

iD8DBQFIGWQO0UIZh3p4ZWERAug/AJ93DWD9o+1VMgPDjWn/dsmPSgTWGQCeOfZi
6LAP26zP25GAeTlKwJQ17hs=
=fwnP
-----END PGP SIGNATURE-----
'''

UNPARSED_PARAGRAPHS_WITH_COMMENTS = '''\
# Leading comments should be ignored.

Source: foo
Section: bar
# An inline comment in the middle of a paragraph should be ignored.
Priority: optional
Homepage: http://www.debian.org/
Build-Depends: debhelper,
# quux, (temporarily disabled by a comment character)
 python

# Comments in the middle shouldn't result in extra blank paragraphs either.

# Ditto.

# A comment at the top of a paragraph should be ignored.
Package: foo
Architecture: any
Description: An awesome package
  # This should still appear in the result.
  Blah, blah, blah. # So should this.
# A comment at the end of a paragraph should be ignored.

# Trailing comments shouldn't cause extra blank paragraphs.
'''

PARSED_PARAGRAPHS_WITH_COMMENTS = [
    deb822.Deb822Dict([
        ('Source', 'foo'),
        ('Section', 'bar'),
        ('Priority', 'optional'),
        ('Homepage', 'http://www.debian.org/'),
        ('Build-Depends', 'debhelper,\n python'),
    ]),
    deb822.Deb822Dict([
        ('Package', 'foo'),
        ('Architecture', 'any'),
        ('Description', 'An awesome package\n'
            '  # This should still appear in the result.\n'
            '  Blah, blah, blah. # So should this.'),
    ]),
]


def find_test_file(filename):
    # type: (str) -> str
    """ find a test file that is located within the test suite """
    return os.path.join(os.path.dirname(__file__), filename)


KEYRING = os.path.abspath(find_test_file('test-keyring.gpg'))


def open_utf8(filename, mode='r'):
    # type: (str, str) -> IO[Text]
    """Open a UTF-8 text file in text mode."""
    return open(filename, mode=mode, encoding='UTF-8')


class TestDeb822Dict:
    def make_dict(self):
        # type: () -> deb822.Deb822Dict
        d = deb822.Deb822Dict()
        d['TestKey'] = 1
        d['another_key'] = 2

        return d

    def test_case_insensitive_lookup(self):
        # type: () -> None
        d = self.make_dict()

        assert 1 == d['testkey']
        assert 2 == d['Another_keY']

    def test_case_insensitive_assignment(self):
        # type: () -> None
        d = self.make_dict()
        d['testkey'] = 3

        assert 3 == d['TestKey']
        assert 3 == d['testkey']

        d.setdefault('foo', 4)
        assert 4 == d['Foo']

    def test_case_preserved(self):
        # type: () -> None
        d = self.make_dict()

        assert sorted(['another_key', 'TestKey']) == sorted(d.keys())

    def test_order_preserved(self):
        # type: () -> None
        d = self.make_dict()
        d['Third_key'] = 3
        d['another_Key'] = 2.5

        keys = ['TestKey', 'another_key', 'Third_key']

        assert keys == list(d.keys())
        assert list(zip(keys, d.values())) == list(d.items())

        keys2 = []
        for key in d:
            keys2.append(key)

        assert keys == keys2

    def test_derived_dict_equality(self):
        # type: () -> None
        d1 = self.make_dict()
        d2 = dict(d1)

        assert d1 == d2

    def test_unicode_key_access(self):
        # type: () -> None
        d = self.make_dict()
        assert 1 == d['testkey']


class TestDeb822:
    def assertWellParsed(self, deb822_, dict_):
        # type: (deb822.Deb822, deb822.Deb822Mapping) -> None
        """Check that the given Deb822 object has the very same keys and
           values as the given dict.
        """

        assert deb822_.keys() == dict_.keys()

        for k, v in dict_.items():
            assert v == deb822_[k]
        assert deb822_ == dict_

    @staticmethod
    def gen_random_string(length=20):
        # type: (int) -> str
        from random import choice
        import string
        chars = string.ascii_letters + string.digits
        return ''.join([choice(chars) for i in range(length)])

    def test_apt_pkg_installed(self):
        # type: () -> None
        # If test suite is running in FORBID_MISSING_APT_PKG mode where
        # python-apt is mandatory, explicitly include a failing test to
        # highlight this problem.
        if FORBID_MISSING_APT_PKG and not _have_apt_pkg:
            pytest.fail("Required apt_pkg from python-apt is not installed (tests run in FORBID_MISSING_APT_PKG mode)")

    def test_gpgv_installed(self):
        # type: () -> None
        # If test suite is running in FORBID_MISSING_GPGV mode where
        # having gpgv is mandatory, explicitly include a failing test to
        # highlight this problem.
        if FORBID_MISSING_GPGV and not _have_gpgv:
            pytest.fail("Required gpgv executable is not installed (tests run in FORBID_MISSING_GPGV mode)")

    def test_parser(self):
        # type: () -> None
        deb822_ = deb822.Deb822(UNPARSED_PACKAGE.splitlines())
        self.assertWellParsed(deb822_, PARSED_PACKAGE)

    def test_pickling(self):
        # type: () -> None
        # Ensure that the Deb822 objects can be serialised
        # See https://bugs.debian.org/975915
        env = os.environ.copy()

        env['PYTHONHASHSEED'] = 'random'
        env['PYTHONPATH'] = os.pathsep.join(sys.path)
        args = [
            sys.executable,
            '-c',
            '''\
from debian import deb822
import pickle
d = deb822.Deb822("Field: value")
with open("test_deb822.pickle", "wb") as fh:
    pickle.dump(d, fh)
        '''
        ]

        try:
            # make the pickle in a different subprocess which will
            # seed the hashes differently
            subprocess.check_call(args, env=env)

            # now read it back
            with open("test_deb822.pickle", "rb") as fh:
                deb822_u = pickle.load(fh)

        finally:
            os.remove("test_deb822.pickle")

        # test that the field could do a round-trip
        assert deb822_u['Field'] == 'value'

    def test_parser_with_newlines(self):
        # type: () -> None
        deb822_ = deb822.Deb822([ l+'\n' for l in UNPARSED_PACKAGE.splitlines()])
        self.assertWellParsed(deb822_, PARSED_PACKAGE)

    def test_strip_initial_blanklines(self):
        # type: () -> None
        deb822_ = deb822.Deb822(['\n'] * 3 + UNPARSED_PACKAGE.splitlines())
        self.assertWellParsed(deb822_, PARSED_PACKAGE)

    def test_reorder(self):
        # type: () -> None
        content = textwrap.dedent("""
        Depends: bar
        Description: some-text
        Architecture: any
        Package: foo
        Recommends: baz
        """)
        paragraph = deb822.Deb822(content.splitlines())
        # Verify the starting state
        assert list(paragraph.keys()) == \
            ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']
        # no op
        paragraph.order_last('Recommends')
        assert list(paragraph.keys()) == \
            ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']
        # no op
        paragraph.order_first('Depends')
        assert list(paragraph.keys()) == \
            ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']

        paragraph.order_first('Package')
        assert list(paragraph.keys()) == \
            ['Package', 'Depends', 'Description', 'Architecture', 'Recommends']

        paragraph.order_last('Description')
        assert list(paragraph.keys()) == \
            ['Package', 'Depends', 'Architecture', 'Recommends', 'Description']

        paragraph.order_after('Recommends', 'Depends')
        assert list(paragraph.keys()) == \
            ['Package', 'Depends', 'Recommends', 'Architecture', 'Description']

        paragraph.order_before('Architecture', 'Depends')
        assert list(paragraph.keys()) == \
            ['Package', 'Architecture', 'Depends', 'Recommends', 'Description']

        with pytest.raises(ValueError):
            paragraph.order_after('Architecture', 'Architecture')
        with pytest.raises(ValueError):
            paragraph.order_before('Architecture', 'Architecture')
        with pytest.raises(KeyError):
            paragraph.order_before('Unknown-Field', 'Architecture')
        with pytest.raises(KeyError):
            paragraph.order_before('Architecture', 'Unknown-Field')

    def test_sort_fields(self):
        # type: () -> None
        content = textwrap.dedent("""
        Depends: bar
        Description: some-text
        Architecture: any
        Package: foo
        Recommends: baz
        """)
        paragraph = deb822.Deb822(content.splitlines())
        # Initial state
        assert list(paragraph.keys()) == \
            ['Depends', 'Description', 'Architecture', 'Package', 'Recommends']

        # Sorting defaults to using name
        paragraph.sort_fields()
        assert list(paragraph.keys()) == \
            ['Architecture', 'Depends', 'Description', 'Package', 'Recommends']

        # Sorting using a key function
        order_table = {
            'package': 1,
            'architecture': 2,
            # Dependencies have the same order
            'depends': 3,
            'recommends': 3,
            # Desc definitely last
            'description': 100,
        }
        paragraph.sort_fields(key=lambda x: order_table[x.lower()])
        assert list(paragraph.keys()) == \
            ['Package', 'Architecture', 'Depends', 'Recommends', 'Description']

    def test_gpg_stripping(self):
        # type: () -> None
        for string in GPG_SIGNED:
            unparsed_with_gpg = string % UNPARSED_PACKAGE
            deb822_ = deb822.Deb822(unparsed_with_gpg.splitlines())
            self.assertWellParsed(deb822_, PARSED_PACKAGE)

    @pytest.mark.skipif(not _have_gpgv, reason="gpgv not installed")
    def test_gpg_info(self):
        # type: () -> None
        unparsed_with_gpg = SIGNED_CHECKSUM_CHANGES_FILE % CHECKSUM_CHANGES_FILE
        deb822_from_str = deb822.Dsc(unparsed_with_gpg)
        result_from_str = deb822_from_str.get_gpg_info(keyrings=[KEYRING])
        deb822_from_file = deb822.Dsc(io.StringIO(unparsed_with_gpg))
        result_from_file = deb822_from_file.get_gpg_info(keyrings=[KEYRING])
        deb822_from_lines = deb822.Dsc(unparsed_with_gpg.splitlines())
        result_from_lines = deb822_from_lines.get_gpg_info(keyrings=[KEYRING])

        valid = {
         'GOODSIG':  ['D14219877A786561', 'John Wright <john.wright@hp.com>'],
         'VALIDSIG': ['8FEFE900783CF175827C2F65D14219877A786561', '2008-05-01',
                      '1209623566', '0', '3', '0', '17', '2', '01',
                      '8FEFE900783CF175827C2F65D14219877A786561'],
         'SIG_ID':   ['j3UjSpdky92fcQISbm8W5PlwC/g', '2008-05-01',
                      '1209623566'],
        }

        for result in result_from_str, result_from_file, result_from_lines:
            # The second part of the GOODSIG field could change if the primary
            # uid changes, so avoid checking that.  Also, the first part of the
            # SIG_ID field has undergone at least one algorithm changein gpg,
            # so don't bother testing that either.
            assert set(result.keys()) == set(valid.keys())
            assert result['GOODSIG'][0] == valid['GOODSIG'][0]
            assert result['VALIDSIG'] == valid['VALIDSIG']
            assert result['SIG_ID'][1:] == valid['SIG_ID'][1:]

    @pytest.mark.skipif(not _have_gpgv, reason="gpgv not installed")
    def test_gpg_info2(self):
        # type: () -> None
        with open(find_test_file('test_Dsc.badsig'), mode='rb') as f:
            dsc = deb822.Dsc(f)
            i = dsc.get_gpg_info(keyrings=[KEYRING])
            assert i.valid()
            assert 'at' == dsc['Source']

    def test_iter_paragraphs_array(self):
        # type: () -> None
        text = (UNPARSED_PACKAGE + '\n\n\n' + UNPARSED_PACKAGE).splitlines()

        for d in deb822.Deb822.iter_paragraphs(text):
            self.assertWellParsed(d, PARSED_PACKAGE)

    def test_iter_paragraphs_file_io(self):
        # type: () -> None
        text = io.StringIO(UNPARSED_PACKAGE + '\n\n\n' + UNPARSED_PACKAGE)

        for d in deb822.Deb822.iter_paragraphs(text, use_apt_pkg=False):
            self.assertWellParsed(d, PARSED_PACKAGE)

    @pytest.mark.skipif(not _have_apt_pkg, reason="apt_pkg is not available")
    def test_iter_paragraphs_file_io_apt_pkg(self):
        # type: () -> None
        text = io.StringIO(UNPARSED_PACKAGE + '\n\n\n' + UNPARSED_PACKAGE)

        with pytest.warns(UserWarning):
            # The io.StringIO is not a real file so this will raise a warning
            for d in deb822.Deb822.iter_paragraphs(text, use_apt_pkg=True):
                self.assertWellParsed(d, PARSED_PACKAGE)

    def test_iter_paragraphs_file(self):
        # type: () -> None
        text = io.StringIO()
        text.write(UNPARSED_PACKAGE)
        text.write('\n\n\n')
        text.write(UNPARSED_PACKAGE)

        with tempfile.NamedTemporaryFile() as fh:
            txt = text.getvalue().encode('UTF-8')
            fh.write(txt)

            fh.seek(0)
            for d in deb822.Deb822.iter_paragraphs(fh, use_apt_pkg=False):
                self.assertWellParsed(d, PARSED_PACKAGE)

    @pytest.mark.skipif(not _have_apt_pkg, reason="apt_pkg is not available")
    def test_iter_paragraphs_file_apt_pkg(self):
        # type: () -> None
        text = io.StringIO()
        text.write(UNPARSED_PACKAGE)
        text.write('\n\n\n')
        text.write(UNPARSED_PACKAGE)

        with tempfile.NamedTemporaryFile() as fh:
            txt = text.getvalue().encode('UTF-8')
            fh.write(txt)

            fh.seek(0)
            for d in deb822.Deb822.iter_paragraphs(fh, use_apt_pkg=True):
                self.assertWellParsed(d, PARSED_PACKAGE)

    def test_iter_paragraphs_with_gpg(self):
        # type: () -> None
        for string in GPG_SIGNED:
            string = string % UNPARSED_PACKAGE
            text = (string + '\n\n\n' + string).splitlines()

            count = 0
            for d in deb822.Deb822.iter_paragraphs(text):
                self.assertWellParsed(d, PARSED_PACKAGE)
                count += 1

            assert 2 == count

    def test_iter_paragraphs_bytes(self):
        # type: () -> None
        text = (UNPARSED_PACKAGE + '\n\n\n' + UNPARSED_PACKAGE)
        binary = text.encode('utf-8')

        for d in deb822.Deb822.iter_paragraphs(binary):
            self.assertWellParsed(d, PARSED_PACKAGE)

    def _test_iter_paragraphs_count(self, filename, cmd, expected, *args, **kwargs):
        # type: (str, Callable[..., Any], int, *Any, **Any) -> None
        with open_utf8(filename) as fh:
            count = len(list(cmd(fh, *args, **kwargs)))
            assert expected == count, \
                "Wrong number paragraphs were found: expected {expected}, got {count}".format(
                    count=count,
                    expected=expected,
                )

    def _test_iter_paragraphs_with_extra_whitespace(self, tests):
        # type: (Callable[[str], None]) -> None
        """ Paragraphs splitting when stray whitespace is between

        From policy §5.1:

            The paragraphs are separated by empty lines. Parsers may accept
            lines consisting solely of spaces and tabs as paragraph separators,
            but control files should use empty lines.

        On the principle of "be strict in what you send; be generous in
        what you receive", deb822 should permit such extra whitespace between
        deb822 stanzas. See #715558 for further details.

        However, when dealing with Packages and Sources files, the behaviour
        of apt is to not split on whitespace-only lines, and so the internal
        parser must be able to avoid doing so when dealing with these data.
        See #913274 for further details.
        """
        for extra_space in (" ", "  ", "\t"):
            text = UNPARSED_PACKAGE + '%s\n' % extra_space + \
                        UNPARSED_PACKAGE

            fd, filename = tempfile.mkstemp()
            fp = os.fdopen(fd, 'wb')
            fp.write(text.encode('utf-8'))
            fp.close()

            try:
                tests(filename)
            finally:
                os.remove(filename)

    def test_iter_paragraphs_with_extra_whitespace_default(self):
        # type: () -> None
        """ Paragraphs splitting with stray whitespace (default options) """
        def tests(filename): # type: (str) -> None
            # apt_pkg not used, should split
            self._test_iter_paragraphs_count(filename, deb822.Deb822.iter_paragraphs, 2)

        self._test_iter_paragraphs_with_extra_whitespace(tests)

    def test_iter_paragraphs_with_extra_whitespace_no_apt_pkg(self):
        # type: () -> None
        """ Paragraphs splitting with stray whitespace (without apt_pkg)"""
        def tests(filename): # type: (str) -> None
            # apt_pkg not used, should split
            self._test_iter_paragraphs_count(filename, deb822.Deb822.iter_paragraphs, 2, use_apt_pkg=False)

            # Specialised iter_paragraphs force use of apt_pkg and don't split
            self._test_iter_paragraphs_count(filename, deb822.Packages.iter_paragraphs, 1, use_apt_pkg=False)
            self._test_iter_paragraphs_count(filename, deb822.Sources.iter_paragraphs, 1, use_apt_pkg=False)

            # Explicitly set internal parser to not split
            strict = {'whitespace-separates-paragraphs': False}
            self._test_iter_paragraphs_count(filename, deb822.Packages.iter_paragraphs, 1, use_apt_pkg=False, strict=strict)
            self._test_iter_paragraphs_count(filename, deb822.Sources.iter_paragraphs, 1, use_apt_pkg=False, strict=strict)

            # Explicitly set internal parser to split
            strict = {'whitespace-separates-paragraphs': True}
            self._test_iter_paragraphs_count(filename, deb822.Packages.iter_paragraphs, 2, use_apt_pkg=False, strict=strict)
            self._test_iter_paragraphs_count(filename, deb822.Sources.iter_paragraphs, 2, use_apt_pkg=False, strict=strict)

        self._test_iter_paragraphs_with_extra_whitespace(tests)

    @pytest.mark.skipif(not _have_apt_pkg, reason="apt_pkg is not available")
    def test_iter_paragraphs_with_extra_whitespace_apt_pkg(self):
        # type: () -> None
        """ Paragraphs splitting with stray whitespace (with apt_pkg) """
        def tests(filename): # type: (str) -> None

            # apt_pkg used, should not split
            self._test_iter_paragraphs_count(filename, deb822.Deb822.iter_paragraphs, 1, use_apt_pkg=True)

            # Specialised iter_paragraphs force use of apt_pkg and don't split
            self._test_iter_paragraphs_count(filename, deb822.Packages.iter_paragraphs, 1, use_apt_pkg=True)
            self._test_iter_paragraphs_count(filename, deb822.Sources.iter_paragraphs, 1, use_apt_pkg=True)

        self._test_iter_paragraphs_with_extra_whitespace(tests)

    def _test_iter_paragraphs(self, filename, cls, **kwargs):
        # type: (str, Type[deb822.Deb822], **Any) -> None
        """Ensure iter_paragraphs consistency"""
        
        with open(filename, 'rb') as fh:
            packages_content = fh.read()

        # XXX: The way multivalued fields parsing works, we can't guarantee
        # that trailing whitespace is reproduced.
        packages_content = b"\n".join([line.rstrip() for line in
                                       packages_content.splitlines()] + [b''])

        s = io.BytesIO()
        l = []
        with warnings.catch_warnings(record=True) as warnings_record:
            with open_utf8(filename) as f:
                for p in cls.iter_paragraphs(f, **kwargs):
                    p.dump(s)
                    s.write(b"\n")
                    l.append(p)
        if not _have_apt_pkg and kwargs.get("use_apt_pkg", True):
            assert len(warnings_record) == 1, "Expected warning not emitted from deb822"
            assert "apt_pkg" in str(warnings_record[0].message)
            assert "python3-apt" in str(warnings_record[0].message)
        assert s.getvalue() == packages_content
        if kwargs["shared_storage"] is False:
            # If shared_storage is False, data should be consistent across
            # iterations -- i.e. we can use "old" objects
            s = io.BytesIO()
            for p in l:
                p.dump(s)
                s.write(b"\n")
            assert s.getvalue() == packages_content

    def test_iter_paragraphs_apt_shared_storage_packages(self):
        # type: () -> None
        self._test_iter_paragraphs(find_test_file("test_Packages"),
                                   deb822.Packages,
                                   use_apt_pkg=True, shared_storage=True)
    def test_iter_paragraphs_apt_no_shared_storage_packages(self):
        # type: () -> None
        self._test_iter_paragraphs(find_test_file("test_Packages"),
                                   deb822.Packages,
                                   use_apt_pkg=True, shared_storage=False)
    def test_iter_paragraphs_no_apt_no_shared_storage_packages(self):
        # type: () -> None
        self._test_iter_paragraphs(find_test_file("test_Packages"),
                                   deb822.Packages,
                                   use_apt_pkg=False, shared_storage=False)

    def test_iter_paragraphs_apt_shared_storage_sources(self):
        # type: () -> None
        self._test_iter_paragraphs(find_test_file("test_Sources"),
                                   deb822.Sources,
                                   use_apt_pkg=True, shared_storage=True)
    def test_iter_paragraphs_apt_no_shared_storage_sources(self):
        # type: () -> None
        self._test_iter_paragraphs(find_test_file("test_Sources"),
                                   deb822.Sources,
                                   use_apt_pkg=True, shared_storage=False)
    def test_iter_paragraphs_no_apt_no_shared_storage_sources(self):
        # type: () -> None
        self._test_iter_paragraphs(find_test_file("test_Sources"),
                                   deb822.Sources,
                                   use_apt_pkg=False, shared_storage=False)

    def test_parser_empty_input(self):
        # type: () -> None
        assert {} == deb822.Deb822([])

    def test_iter_paragraphs_empty_input(self):
        # type: () -> None
        generator = deb822.Deb822.iter_paragraphs([])
        with pytest.raises(StopIteration):
            next(generator)

    def test_parser_limit_fields(self):
        # type: () -> None
        wanted_fields = [ 'Package', 'MD5sum', 'Filename', 'Description' ]
        deb822_ = deb822.Deb822(UNPARSED_PACKAGE.splitlines(), wanted_fields)

        assert sorted(wanted_fields) == sorted(deb822_.keys())

        for key in wanted_fields:
            assert PARSED_PACKAGE[key] == deb822_[key]

    def test_iter_paragraphs_limit_fields(self):
        # type: () -> None
        wanted_fields = [ 'Package', 'MD5sum', 'Filename', 'Tag' ]

        for deb822_ in deb822.Deb822.iter_paragraphs(
                UNPARSED_PACKAGE.splitlines(), wanted_fields):

            assert sorted(wanted_fields) == sorted(deb822_.keys())

            for key in wanted_fields:
                assert PARSED_PACKAGE[key] == deb822_[key]

    def test_dont_assume_trailing_newline(self):
        # type: () -> None
        deb822a = deb822.Deb822(['Package: foo'])
        deb822b = deb822.Deb822(['Package: foo\n'])

        assert deb822a['Package'] == deb822b['Package']

        deb822a = deb822.Deb822(['Description: foo\n', 'bar'])
        deb822b = deb822.Deb822(['Description: foo', 'bar\n'])

        assert deb822a['Description'] == deb822b['Description']

    def test__delitem__(self):
        # type: () -> None
        parsed = deb822.Deb822(UNPARSED_PACKAGE.splitlines())
        deriv = deb822.Deb822(_parsed=parsed)
        dict_ = PARSED_PACKAGE.copy()

        for key in ('Package', 'MD5sum', 'Description'):
            del dict_[key]
            for d in (parsed, deriv):
                del d[key]
                d.keys() # ensure this does not raise error
                self.assertWellParsed(d, dict_)


    def test_policy_compliant_whitespace(self):
        # type: () -> None
        string = (
            'Package: %(Package)s\n'
            'Version :%(Version)s \n'
            'Priority:%(Priority)s\t \n'
            'Section \t :%(Section)s \n'
            'Empty-Field:        \t\t\t\n'
            'Multiline-Field : a \n b\n c\n'
        ) % PARSED_PACKAGE

        deb822_ = deb822.Deb822(string.splitlines())
        dict_   = PARSED_PACKAGE.copy()

        dict_['Empty-Field'] = ''
        dict_['Multiline-Field'] = 'a\n b\n c' # XXX should be 'a\nb\nc'?

        for k, v in deb822_.items():
            assert dict_[k] == v
    
    def test_case_insensitive(self):
        # type: () -> None
        # PARSED_PACKAGE is a deb822.Deb822Dict object, so we can test
        # it directly
        assert PARSED_PACKAGE['Architecture'] == PARSED_PACKAGE['architecture']

        c_i_dict = deb822.Deb822Dict()

        test_string = self.gen_random_string()
        c_i_dict['Test-Key'] = test_string
        assert test_string == c_i_dict['test-key']

        test_string_2 = self.gen_random_string()
        c_i_dict['TeSt-KeY'] = test_string_2
        assert test_string_2 == c_i_dict['Test-Key']

        deb822_ = deb822.Deb822(io.StringIO(UNPARSED_PACKAGE))
        # deb822_.keys() will return non-normalized keys
        for k in deb822_:
            assert deb822_[k] == deb822_[k.lower()]

    def test_multiline_trailing_whitespace_after_colon(self):
        # type: () -> None
        """Trailing whitespace after the field name on multiline fields

        If the field's value starts with a newline (e.g. on MD5Sum fields in
        Release files, or Files field in .dsc's, the dumped string should not
        have a trailing space after the colon.  If the value does not start
        with a newline (e.g. the control file Description field), then there
        should be a space after the colon, as with non-multiline fields.
        """
        
        # bad_re: match a line that starts with a "Field:", and ends in
        # whitespace
        bad_re = re.compile(r"^\S+:\s+$")
        for cls in deb822.Deb822, deb822.Changes:
            parsed = cls(CHANGES_FILE.splitlines())
            for line in parsed.dump().splitlines():
                assert bad_re.match(line) is None, \
                                "There should not be trailing whitespace " \
                                "after the colon in a multiline field " \
                                "starting with a newline"

        
        control_paragraph = """Package: python-debian
Architecture: all
Depends: ${python:Depends}
Suggests: python-apt
Provides: python-deb822
Conflicts: python-deb822
Replaces: python-deb822
Description: python modules to work with Debian-related data formats
 This package provides python modules that abstract many formats of Debian
 related files. Currently handled are:
  * Debtags information (debian_bundle.debtags module)
  * debian/changelog (debian_bundle.changelog module)
  * Packages files, pdiffs (debian_bundle.debian_support module)
  * Control files of single or multple RFC822-style paragraphs, e.g
    debian/control, .changes, .dsc, Packages, Sources, Release, etc.
    (debian_bundle.deb822 module)
"""
        parsed_control = deb822.Deb822(control_paragraph.splitlines())
        field_re = re.compile(r"^\S+:")
        field_with_space_re = re.compile(r"^\S+: ")
        dump = parsed_control.dump()
        assert dump is not None
        for line in dump.splitlines():
            if field_re.match(line):
                assert field_with_space_re.match(line), \
                                "Multiline fields that do not start with " \
                                "newline should have a space between the " \
                                "colon and the beginning of the value"

    def test_blank_value(self):
        # type: () -> None
        """Fields with blank values are parsable--so they should be dumpable"""

        d = deb822.Deb822()
        d['Foo'] = 'bar'
        d['Baz'] = ''
        d['Another-Key'] = 'another value'
        
        # Previous versions would raise an exception here -- this makes the
        # test fail and gives useful information, so I won't try to wrap around
        # it.
        dumped = d.dump()
        
        # May as well make sure the resulting string is what we want
        expected = "Foo: bar\nBaz:\nAnother-Key: another value\n"
        assert dumped == expected

    def test_copy(self):
        # type: () -> None
        """The copy method of Deb822 should return another Deb822 object"""
        d = deb822.Deb822()
        d['Foo'] = 'bar'
        d['Bar'] = 'baz'
        d_copy = d.copy()

        assert isinstance(d_copy, deb822.Deb822)
        expected_dump = "Foo: bar\nBar: baz\n"
        assert d_copy.dump() == expected_dump

    def test_bug457929_multivalued_dump_works(self):
        # type: () -> None
        """dump() was not working in multivalued classes, see #457929."""
        changesobj = deb822.Changes(CHANGES_FILE.splitlines())
        assert CHANGES_FILE == changesobj.dump()

    def test_bug487902_multivalued_checksums(self):
        # type: () -> None
        """New multivalued field Checksums was not handled correctly, see #487902."""
        changesobj = deb822.Changes(CHECKSUM_CHANGES_FILE.splitlines())
        assert CHECKSUM_CHANGES_FILE == changesobj.dump()

    def test_case_preserved_in_input(self):
        # type: () -> None
        """The field case in the output from dump() should be the same as the
        input, even if multiple Deb822 objects have been created using
        different case conventions.

        This is related to bug 473254 - the fix for this issue is probably the
        same as the fix for that bug.
        """
        input1 = "Foo: bar\nBaz: bang\n"
        input2 = "foo: baz\nQux: thud\n"
        d1 = deb822.Deb822(input1.splitlines())
        d2 = deb822.Deb822(input2.splitlines())
        assert input1 == d1.dump()
        assert input2 == d2.dump()

        d3 = deb822.Deb822()
        if 'some-test-key' not in d3:
            d3['Some-Test-Key'] = 'some value'
        assert d3.dump() == "Some-Test-Key: some value\n"

    @pytest.mark.skipif(not _have_apt_pkg, reason="apt_pkg is not available")
    def test_unicode_values_apt_pkg(self):
        # type: () -> None
        """Deb822 objects should contain only unicode values

        (Technically, they are allowed to contain any type of object, but when
        parsed from files, and when only string-type objects are added, the
        resulting object should have only unicode values.)
        """

        objects = []  # type: List[deb822.Deb822]
        with open_utf8(find_test_file('test_Packages')) as f:
            objects.extend(deb822.Packages.iter_paragraphs(f))
        with open_utf8(find_test_file('test_Sources')) as f:
            objects.extend(deb822.Deb822.iter_paragraphs(f))
        with open(find_test_file('test_Sources.iso8859-1'), 'rb') as fb:
            objects.extend(deb822.Deb822.iter_paragraphs(
                fb, encoding="iso8859-1"))
        for d in objects:
            for value in d.values():
                assert isinstance(value, str)

        # The same should be true for Sources and Changes except for their
        # _multivalued fields
        multi = []   # type: List[Union[deb822.Changes, deb822.Sources]]
        with open_utf8(find_test_file('test_Sources')) as f:
            multi.extend(deb822.Sources.iter_paragraphs(f))
        for d in multi:
            for key, value in d.items():
                if key.lower() not in d.__class__._multivalued_fields:
                    assert isinstance(value, str)

    def test_unicode_values(self):
        # type: () -> None
        """Deb822 objects should contain only unicode values

        (Technically, they are allowed to contain any type of object, but when
        parsed from files, and when only string-type objects are added, the
        resulting object should have only unicode values.)
        """

        objects = []
        objects.append(deb822.Deb822(UNPARSED_PACKAGE))
        objects.append(deb822.Deb822(CHANGES_FILE))
        with open_utf8(find_test_file('test_Packages')) as f:
            objects.extend(deb822.Deb822.iter_paragraphs(f))
        with open_utf8(find_test_file('test_Sources')) as f:
            objects.extend(deb822.Deb822.iter_paragraphs(f))
        for d in objects:
            for value in d.values():
                assert isinstance(value, str)

        # The same should be true for Sources and Changes except for their
        # _multivalued fields
        multi = []   # type: List[Union[deb822.Changes, deb822.Sources]]
        multi.append(deb822.Changes(CHANGES_FILE))
        multi.append(deb822.Changes(SIGNED_CHECKSUM_CHANGES_FILE
                                    % CHECKSUM_CHANGES_FILE))
        for d in multi:
            for key, value in d.items():
                if key.lower() not in d.__class__._multivalued_fields:
                    assert isinstance(value, str)

    def test_encoding_integrity(self):
        # type: () -> None
        with open_utf8(find_test_file('test_Sources')) as f:
            utf8 = list(deb822.Deb822.iter_paragraphs(f))
        with open(find_test_file('test_Sources.iso8859-1'), 'rb') as fb:
            latin1 = list(deb822.Deb822.iter_paragraphs(
                fb, encoding='iso8859-1'))

        # dump() with no fd returns a unicode object - both should be identical
        assert len(utf8) == len(latin1)
        for i in range(len(utf8)):
            assert utf8[i].dump() == latin1[i].dump()

        # XXX: The way multiline fields parsing works, we can't guarantee
        # that trailing whitespace is reproduced.
        with open(find_test_file('test_Sources'), 'rb') as fb:
            utf8_contents = b"\n".join([line.rstrip() for line in fb] + [b''])
        with open(find_test_file('test_Sources.iso8859-1'), 'rb') as fb:
            latin1_contents = b"\n".join([line.rstrip() for line in fb] + [b''])

        utf8_to_latin1 = io.BytesIO()
        for d in utf8:
            d.dump(fd=utf8_to_latin1, encoding='iso8859-1')
            utf8_to_latin1.write(b"\n")

        latin1_to_utf8 = io.BytesIO()
        for d in latin1:
            d.dump(fd=latin1_to_utf8, encoding='utf-8')
            latin1_to_utf8.write(b"\n")

        assert utf8_contents == latin1_to_utf8.getvalue()
        assert latin1_contents == utf8_to_latin1.getvalue()

    def test_mixed_encodings(self):
        # type: () -> None
        """Test that we can handle a simple case of mixed encodings

        In general, this isn't guaranteed to work.  It uses the chardet
        package, which tries to determine heuristically the encoding of the
        text given to it.  But as far as I've seen, it's reliable for mixed
        latin1 and utf-8 in maintainer names in old Sources files...
        """

        # Avoid spitting out the encoding warning during testing.
        warnings.filterwarnings(action='ignore', category=UnicodeWarning)

        filename = find_test_file('test_Sources.mixed_encoding')
        with open(filename, 'rb') as f1, open(filename, 'rb') as f2:
            for paragraphs in [
                        deb822.Sources.iter_paragraphs(f1),
                        deb822.Sources.iter_paragraphs(f2, use_apt_pkg=False)
                    ]:
                with warnings.catch_warnings(record=True) as warnings_record:
                    p1 = next(paragraphs)
                    assert p1['maintainer'] == \
                                    'Adeodato Sim\xf3 <dato@net.com.org.es>'
                    p2 = next(paragraphs)
                    assert p2['uploaders'] == \
                                    'Frank K\xfcster <frank@debian.org>'
                if FORBID_MISSING_APT_PKG:
                    assert not warnings_record, "Warnings emitted from deb822"

    def test_dump_text_mode(self):
        # type: () -> None
        d = deb822.Deb822(CHANGES_FILE.splitlines())
        buf = io.StringIO()
        d.dump(fd=buf, text_mode=True)
        assert CHANGES_FILE == buf.getvalue()


    def test_bug597249_colon_as_first_value_character(self):
        # type: () -> None
        """Colon should be allowed as the first value character. See #597249.
        """

        data = 'Foo: : bar'
        parsed = {'Foo': ': bar'}
        self.assertWellParsed(deb822.Deb822(data), parsed)

    @staticmethod
    def _dictset(d, key, value):
        # type: (Dict[str, Any], str, Any) -> None
        d[key] = value

    def test_field_value_ends_in_newline(self):
        # type: () -> None
        """Field values are not allowed to end with newlines"""

        d = deb822.Deb822()
        with pytest.raises(ValueError):
            self._dictset(d, 'foo', 'bar\n')   # type: ignore
        with pytest.raises(ValueError):
            self._dictset(d, 'foo', 'bar\nbaz\n')  # type: ignore

    def test_field_value_contains_blank_line(self):
        # type: () -> None
        """Field values are not allowed to contain blank lines"""

        d = deb822.Deb822()
        with pytest.raises(ValueError):
            self._dictset(d, 'foo', 'bar\n\nbaz')  # type: ignore
        with pytest.raises(ValueError):
            self._dictset(d, 'foo', '\n\nbaz')  # type: ignore

    def test_multivalued_field_contains_newline(self):
        # type: () -> None
        # mypy does not handle metaprogramming for multivalued fields
        """Multivalued field components are not allowed to contain newlines"""

        d = deb822.Dsc()
        # We don't check at set time, since one could easily modify the list
        # without deb822 knowing.  We instead check at get time.
        d['Files'] = [{'md5sum': 'deadbeef', 'size': '9605', 'name': 'bad\n'}]   # type: ignore
        with pytest.raises(ValueError):
            d.get_as_string('files')

    def _test_iter_paragraphs_comments(self, paragraphs):
        # type: (List[deb822.Deb822]) -> None
        assert len(paragraphs) == len(PARSED_PARAGRAPHS_WITH_COMMENTS)
        for i in range(len(paragraphs)):
            self.assertWellParsed(paragraphs[i],
                                  PARSED_PARAGRAPHS_WITH_COMMENTS[i])

    @pytest.mark.skipif(not _have_apt_pkg, reason="apt_pkg is not available")
    def test_iter_paragraphs_comments_use_apt_pkg(self):
        # type: () -> None
        """ apt_pkg does not support comments within multiline fields

        This test checks that a file with comments inside multiline fields
        generates an error from the apt_pkg parser.

        See also https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=750247#35
                 https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=807351
        """
        try:
            fd, filename = tempfile.mkstemp()
            fp = os.fdopen(fd, 'wb')
            fp.write(UNPARSED_PARAGRAPHS_WITH_COMMENTS.encode('utf-8'))
            fp.close()

            with open_utf8(filename) as fh:
                with pytest.raises(apt_pkg.Error):
                    list(deb822.Deb822.iter_paragraphs(fh, use_apt_pkg=True))
        finally:
            os.remove(filename)

    def test_iter_paragraphs_comments_native(self):
        # type: () -> None
        paragraphs = list(deb822.Deb822.iter_paragraphs(
            UNPARSED_PARAGRAPHS_WITH_COMMENTS.splitlines(), use_apt_pkg=False))
        self._test_iter_paragraphs_comments(paragraphs)

    def test_iter_paragraphs_string_comments_native(self):
        # type: () -> None
        paragraphs = list(deb822.Deb822.iter_paragraphs(
            UNPARSED_PARAGRAPHS_WITH_COMMENTS, use_apt_pkg=False))
        self._test_iter_paragraphs_comments(paragraphs)

    def test_explicit_source_field(self):
        # type: () -> None
        """Test handling of explicit Source field in Packages """

        # implicit source and version
        pkg = next(deb822.Packages.iter_paragraphs(
            UNPARSED_PACKAGE,
            use_apt_pkg=False,
        ))
        assert 'mutt' == pkg.source
        assert '1.5.12-1' == str(pkg.source_version)
        assert isinstance(pkg.source_version, Version)

        # explicit source, implicit version
        pkg = next(deb822.Packages.iter_paragraphs(
            UNPARSED_PACKAGE + "Source: mutt-dummy\n",
            use_apt_pkg=False,
        ))
        assert 'mutt-dummy' == pkg.source
        assert '1.5.12-1' == str(pkg.source_version)
        assert isinstance(pkg.source_version, Version)

        pkg = next(deb822.Packages.iter_paragraphs(
            UNPARSED_PACKAGE + "Source: mutt-dummy (1.2.3-1)\n",
            use_apt_pkg=False,
        ))
        assert 'mutt-dummy' == pkg.source
        assert '1.2.3-1' == str(pkg.source_version)
        assert isinstance(pkg.source_version, Version)

    def test_release(self):
        # type: () -> None
        with open(find_test_file('test_Release')) as f:
            release = deb822.Release(f)
        assert release['Codename'] == 'sid'
        assert len(release['SHA1']) == 61
        assert len(release['SHA256']) == 61
        assert len(release['SHA512']) == 61
        assert release['SHA512'][0]['size'] == '113433'

    def test_buildinfo(self):
        # type: () -> None
        with open(find_test_file('test_BuildInfo')) as f:
            buildinfo = deb822.BuildInfo(f)
        assert buildinfo['Build-Origin'] == 'Debian'
        assert 'build-essential' in buildinfo['Installed-Build-Depends']
        ibd = buildinfo.relations['installed-build-depends']
        assert len(ibd) == 207  # type: ignore

        benv = buildinfo.get_environment()
        assert len(benv) == 3
        assert benv['DEB_BUILD_OPTIONS'] == 'parallel=4'

        ch = buildinfo.get_changelog()
        assert ch is not None
        assert ch.version == '4.5.2-1+b1'
        assert ch.package == 'lxml'
        assert ch.urgency == 'low'
        assert 'Daemon' in ch.author

        del buildinfo['Binary-Only-Changes']
        ch = buildinfo.get_changelog()
        assert ch is None

    def test_buildinfo_env_deserialise(self):
        # type: () -> None
        data = r"""
 DEB_BUILD_OPTIONS="parallel=4"
 LANG="en_AU.UTF-8"
 LC_ALL="C
UTF-8"
 LC_TIME="en_GB.UTF-8"
 LD_LIBRARY_PATH="/usr/lib/libeatmy\"data"
 SOURCE_DATE_EPOCH="1601784586"
"""
        benv = dict(deb822.BuildInfo._env_deserialise(data))

        assert len(benv) == 6
        assert benv['DEB_BUILD_OPTIONS'] == 'parallel=4'
        assert '\n' in benv['LC_ALL']
        assert '"' in benv['LD_LIBRARY_PATH']
        assert benv['LD_LIBRARY_PATH'] == '/usr/lib/libeatmy"data'

        benv = dict(deb822.BuildInfo._env_deserialise(""))

        assert len(benv) == 0

    def test_changes_binary_mode(self):
        # type: () -> None
        """Trivial parse test for a signed file in binary mode"""
        with io.open(find_test_file('test_Changes'), 'rb') as f:
            changes = deb822.Changes(f)
        assert 'python-debian' == changes['Source']

    def test_changes_text_mode(self):
        # type: () -> None
        """Trivial parse test for a signed file in text mode"""
        with io.open(find_test_file('test_Changes'), 'r', encoding='utf-8') as f:
            changes = deb822.Changes(f)
        assert 'python-debian' == changes['Source']

    def test_removals(self):
        # type: () -> None
        with open(find_test_file('test_removals.822')) as f:
            removals = deb822.Removals.iter_paragraphs(f)
            r = next(removals)
            assert r['suite'] == 'unstable'
            assert r['date'] == u'Wed, 01 Jan 2014 17:03:54 +0000'
            # Date objects, timezones, cross-platform, portability nightmare...
            assert r.date.strftime('%S') == '54'
            assert len(r.binaries) == 1
            assert r.binaries[0]['package'] == 'libzoom-ruby'
            assert r.binaries[0]['version'] == '0.4.1-5'
            assert r.binaries[0]['architectures'] == set(['all'])
            r = next(removals)
            assert len(r.binaries) == 3
            r = next(removals)
            assert r['bug'] == '753912'
            assert r.bug == [753912]
            assert r.also_wnpp == [123456]
            r = next(removals)
            assert r.binaries[0]['architectures'] == \
                set(['amd64', 'armel', 'armhf', 'hurd-i386', 'i386',
                     'kfreebsd-amd64', 'kfreebsd-i386', 'mips', 'mipsel',
                     'powerpc', 's390x', 'sparc'])


class TestPkgRelations:

    def assertPkgDictEqual(self, expected, actual):
        # type: (deb822.Deb822Mapping, deb822.Deb822Mapping) -> None
        p1keys = sorted(expected.keys())
        p2keys = sorted(actual.keys())
        assert p1keys == p2keys, "Different fields present in packages"
        for k in p1keys:
            assert expected[k] == actual[k], "Different for field '%s'" % k

    @staticmethod
    def rel(dict_):
        # type: (deb822.Deb822MutableMapping) -> deb822.Deb822Mapping
        """Modify dict_ to ensure it contains all fields from parse_relations

        Accept a dict that partially describes a package relationship and add
        to it any missing keys.

        returns: modified dict_
        """
        if 'version' not in dict_:
            dict_['version'] = None
        if 'arch' not in dict_:
            dict_['arch'] = None
        if 'archqual' not in dict_:
            dict_['archqual'] = None
        if 'restrictions' not in dict_:
            dict_['restrictions'] = None
        return dict_

    def test_packages(self):
        # type: () -> None
        # make the syntax a bit more compact
        rel = TestPkgRelations.rel

        with warnings.catch_warnings(record=True) as warnings_record:
            f = open(find_test_file('test_Packages'))
            pkgs = deb822.Packages.iter_paragraphs(f)
            pkg1 = next(pkgs)
            rel1 = {'breaks': [],
                    'built-using': [],
                    'conflicts': [],
                    'depends': [
                            [rel({'name': 'file', 'archqual': 'i386'})],
                            [rel({'name': 'libc6', 'version': ('>=', '2.7-1')})],
                            [rel({'name': 'libpaper1'})],
                            [rel({'name': 'psutils'})],
                        ],
                    'enhances': [],
                    'pre-depends': [],
                    'provides': [],
                    'recommends': [
                            [rel({'name': 'bzip2'})],
                            [rel({'name': 'lpr'}),
                                rel({'name': 'rlpr'}),
                                rel({'name': 'cupsys-client'})],
                            [rel({'name': 'wdiff'})],
                        ],
                    'replaces': [],
                    'suggests': [
                            [rel({'name': 'emacsen-common'})],
                            [rel({'name': 'ghostscript'})],
                            [rel({'name': 'graphicsmagick-imagemagick-compat'}),
                                rel({'name': 'imagemagick'})],
                            [rel({'name': 'groff'})],
                            [rel({'name': 'gv'})],
                            [rel({'name': 'html2ps'})],
                            [rel({'name': 't1-cyrillic'})],
                            [rel({'name': 'texlive-base-bin'})],
                        ]
                    }
            self.assertPkgDictEqual(rel1, pkg1.relations)
            pkg2 = next(pkgs)
            rel2 = {'breaks': [],
                    'built-using': [],
                    'conflicts': [],
                    'depends': [
                            [rel({'name': 'lrzsz'})],
                            [rel({'name': 'openssh-client'}),
                                rel({'name': 'telnet'}),
                                rel({'name': 'telnet-ssl'})],
                            [rel({'name': 'libc6', 'version': ('>=', '2.6.1-1')})],
                            [rel({'name': 'libncurses5', 'version': ('>=', '5.6')})],
                            [rel({'name': 'libreadline5', 'version': ('>=', '5.2')})],
                        ],
                    'enhances': [],
                    'pre-depends': [],
                    'provides': [],
                    'recommends': [],
                    'replaces': [],
                    'suggests': []
                    }
            self.assertPkgDictEqual(rel2, pkg2.relations)
            pkg3 = next(pkgs)
            dep3 = [
                    [rel({'name': 'dcoprss', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kdenetwork-kfile-plugins', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kdict', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kdnssd', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kget', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'knewsticker', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kopete', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kpf', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kppp', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'krdc', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'krfb', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'ksirc', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'kwifimanager', 'version': ('>=', '4:3.5.9-2')})],
                    [rel({'name': 'librss1', 'version': ('>=', '4:3.5.9-2')})],
                ]
            assert dep3 == pkg3.relations['depends']
            f.close()
        if FORBID_MISSING_APT_PKG:
            # Don't permit this test to succeed if warnings about apt_pkg
            # being missing were generated
            assert not warnings_record, "Warnings emitted from deb822"

    def test_pkgrelation_str(self, caplog):
        # type: (pytest.LogCaptureFixture) -> None
        bin_rels = [
            'file, libc6 (>= 2.7-1), libpaper1, psutils, '
            'perl:any, python:native'
            ]
        src_rels = [
            'apache2-src (>= 2.2.9), libaprutil1-dev, '
            'libcap-dev [!kfreebsd-i386 !kfreebsd-amd64 !hurd-i386], '
            'autoconf <!cross>, '
            'debhelper (>> 5.0.0) <!stage1> <!cross !stage2>'
            ]
        for bin_rel in bin_rels:
            assert bin_rel == \
                    deb822.PkgRelation.str(deb822.PkgRelation.parse_relations(
                            bin_rel))
        for src_rel in src_rels:
            assert src_rel == \
                    deb822.PkgRelation.str(deb822.PkgRelation.parse_relations( \
                            src_rel))
        with caplog.at_level(logging.WARNING):
            deb822.PkgRelation.parse_relations("foo bar")
        assert caplog.record_tuples == [(
            "debian.deb822",
            logging.WARNING,
            'cannot parse package relationship '
            '"foo bar", returning it raw'
        )]

    def test_sources(self):
        # type: () -> None
        # make the syntax a bit more compact
        rel = TestPkgRelations.rel

        # Should not get warnings about missing python-apt from this code
        with warnings.catch_warnings(record=True) as warnings_record:
            f = open_utf8(find_test_file('test_Sources'))
            pkgs = deb822.Sources.iter_paragraphs(f)
            pkg1 = next(pkgs)
            rel1 = {'build-conflicts': [],
                    'build-conflicts-indep': [],
                    'build-conflicts-arch': [],
                    'build-depends': [
                            [rel({'name': 'apache2-src', 'version': ('>=', '2.2.9')})],
                            [rel({'name': 'libaprutil1-dev'})],
                            [rel({'arch': [(False, 'kfreebsd-i386'), (False, 'kfreebsd-amd64'), (False, 'hurd-i386')],
                                'name': 'libcap-dev'})],
                            [rel({'name': 'autoconf'})],
                            [rel({'name': 'debhelper', 'version': ('>>', '5.0.0')})],
                        ],
                    'build-depends-indep': [],
                    'build-depends-arch': [],
                    'binary': [
                            [rel({'name': 'apache2-mpm-itk'})]
                        ]
                    }
            self.assertPkgDictEqual(rel1, pkg1.relations)
            pkg2 = next(pkgs)
            rel2 = {'build-conflicts': [],
                    'build-conflicts-indep': [],
                    'build-conflicts-arch': [],
                    'build-depends': [
                            [rel({'name': 'dpkg-dev', 'version': ('>=', '1.13.9')})],
                            [rel({'name': 'autoconf', 'version': ('>=', '2.13')})],
                            [rel({'name': 'bash'})],
                            [rel({'name': 'bison', 'archqual': 'amd64'})],
                            [rel({'name': 'flex'})],
                            [rel({'name': 'gettext', 'archqual': 'any'})],
                            [rel({'name': 'texinfo',
                                'restrictions': [
                                    [(False, 'stage1')],
                                    [(False, 'stage2'),
                                    (False, 'cross')]
                                ]})],
                            [rel({'arch': [(True, 'hppa')], 'name': 'expect-tcl8.3',
                                'version': ('>=', '5.32.2'),
                                'restrictions': [[(False, 'stage1')]]})],
                            [rel({'name': 'dejagnu', 'version': ('>=', '1.4.2-1.1'), 'arch': None})],
                            [rel({'name': 'dpatch'})],
                            [rel({'name': 'file'})],
                            [rel({'name': 'bzip2', 'archqual': 'native'})],
                            [rel({'name': 'lsb-release'})],
                        ],
                    'build-depends-indep': [],
                    'build-depends-arch': [],
                    'binary': [
                            [rel({'name': 'binutils'})],
                            [rel({'name': 'binutils-dev'})],
                            [rel({'name': 'binutils-multiarch'})],
                            [rel({'name': 'binutils-hppa64'})],
                            [rel({'name': 'binutils-spu'})],
                            [rel({'name': 'binutils-doc'})],
                            [rel({'name': 'binutils-source'})],
                        ]
                    }
            self.assertPkgDictEqual(rel2, pkg2.relations)
            f.close()
        if FORBID_MISSING_APT_PKG:
            # Don't permit this test to succeed if warnings about apt_pkg
            # being missing were generated
            assert not warnings_record, "Warnings emitted from deb822"

    def test_restrictions_parse(self):   # type: ignore
        """ test parsing of restriction formulas """
        r = "foo <cross>"
        # relation 0, alternative 0, restrictions set 0, condition 0
        terms = deb822.PkgRelation.parse_relations(r)[0][0]['restrictions']
        assert terms is not None
        term  = terms[0][0]
        assert term.enabled == True
        assert term[0] == True
        assert term.profile == 'cross'
        assert term[1] == 'cross'

        r = "foo <!stage1> <!stage2 !cross>"
        # relation 0, alternative 0, restrictions set 1, condition 0
        terms = deb822.PkgRelation.parse_relations(r)[0][0]['restrictions']
        assert terms is not None
        term = terms[1][0]
        assert term.enabled == False
        assert term[0] == False
        assert term.profile == 'stage2'
        assert term[1] == 'stage2'

        # relation 0, alternative 0, restrictions set 1, condition 1
        terms = deb822.PkgRelation.parse_relations(r)[0][0]['restrictions']
        assert terms is not None
        term = terms[1][1]
        assert term.enabled == False
        assert term[0] == False
        assert term.profile == 'cross'
        assert term[1] == 'cross'

    def test_multiarch_parse(self):
        # type: () -> None
        """ test parsing of architecture qualifiers from multiarch

        Also ensure that the archqual part makes a round-trip, see
        https://bugs.debian.org/868249
        """
        r = "foo:native"
        # relation 0, alternative 0, arch qualifier
        rel = deb822.PkgRelation.parse_relations(r)
        term = rel[0][0]['archqual']
        assert term == "native"
        assert deb822.PkgRelation.str(rel) == r


class TestVersionAccessor:

    def test_get_version(self):
        # type: () -> None
        # should not be available in most basic Deb822
        p = deb822.Deb822(UNPARSED_PACKAGE.splitlines())
        with pytest.raises(AttributeError):
            p.get_version()    # type: ignore

        # should be available in Packages
        p = deb822.Packages(UNPARSED_PACKAGE.splitlines())
        v = p.get_version()
        assert str(v) == '1.5.12-1'
        assert isinstance(v, Version)

    def test_set_version(self):
        # type: () -> None
        # should not be available in most basic Deb822
        p = deb822.Deb822(UNPARSED_PACKAGE.splitlines())
        with pytest.raises(AttributeError):
            p.set_version()   # type: ignore

        # should be available in Packages
        p = deb822.Packages(UNPARSED_PACKAGE.splitlines())
        newver = '9.8.7-1'
        v = Version(newver)
        p.set_version(v)
        assert p['Version'] == newver
        assert isinstance(p['Version'], str)


@pytest.mark.skipif(not _have_gpgv, reason="gpgv not installed")
class TestGpgInfo:

    SampleData = namedtuple('SampleData', [
        'datastr',
        'data',
        'valid',
    ])

    @pytest.fixture()
    def sampledata(self):
        # type: () -> Generator[TestGpgInfo.SampleData, None, None]
        datastr = SIGNED_CHECKSUM_CHANGES_FILE % CHECKSUM_CHANGES_FILE
        data = datastr.encode()
        valid = {
            'GOODSIG':
                ['D14219877A786561', 'John Wright <john.wright@hp.com>'],
            'VALIDSIG':
                ['8FEFE900783CF175827C2F65D14219877A786561', '2008-05-01',
                 '1209623566', '0', '3', '0', '17', '2', '01',
                 '8FEFE900783CF175827C2F65D14219877A786561'],
            'SIG_ID':
                ['j3UjSpdky92fcQISbm8W5PlwC/g', '2008-05-01', '1209623566'],
        }
        yield TestGpgInfo.SampleData(
            datastr,
            data,
            valid,
        )

    def _validate_gpg_info(self, gpg_info, sampledata):
        # type: (deb822.GpgInfo, TestGpgInfo.SampleData) -> None
        # The second part of the GOODSIG field could change if the primary
        # uid changes, so avoid checking that.  Also, the first part of the
        # SIG_ID field has undergone at least one algorithm change in gpg,
        # so don't bother testing that either.
        assert set(gpg_info.keys()) == set(sampledata.valid.keys())
        assert gpg_info['GOODSIG'][0] == sampledata.valid['GOODSIG'][0]
        assert gpg_info['VALIDSIG'] == sampledata.valid['VALIDSIG']
        assert gpg_info['SIG_ID'][1:] == sampledata.valid['SIG_ID'][1:]

    def test_from_sequence_string(self, sampledata):
        # type: (TestGpgInfo.SampleData) -> None
        gpg_info = deb822.GpgInfo.from_sequence(sampledata.data, keyrings=[KEYRING])
        self._validate_gpg_info(gpg_info, sampledata)

    def test_from_sequence_newline_terminated(self, sampledata):
        # type: (TestGpgInfo.SampleData) -> None
        sequence = io.BytesIO(sampledata.data)
        gpg_info = deb822.GpgInfo.from_sequence(sequence, keyrings=[KEYRING])
        self._validate_gpg_info(gpg_info, sampledata)

    def test_from_sequence_no_newlines(self, sampledata):
        # type: (TestGpgInfo.SampleData) -> None
        sequence = sampledata.data.splitlines()
        gpg_info = deb822.GpgInfo.from_sequence(sequence, keyrings=[KEYRING])
        self._validate_gpg_info(gpg_info, sampledata)

    def test_from_file(self, sampledata):
        # type: (TestGpgInfo.SampleData) -> None
        fd, filename = tempfile.mkstemp()
        fp = os.fdopen(fd, 'wb')
        fp.write(sampledata.data)
        fp.close()

        try:
            gpg_info = deb822.GpgInfo.from_file(filename, keyrings=[KEYRING])
        finally:
            os.remove(filename)

        self._validate_gpg_info(gpg_info, sampledata)
